#!/usr/bin/env python3

# A wrapper script for the touch detection algorithm
#
# Usage:
#
# snudda init <networkPath> --size XXX
# -- Creates an a json config file
#
# snudda place <networkPath> [--parallel]
# -- Cell placement within volumes specified
#
# snudda detect <networkPath> [--hvsize hyperVoxelSize] [--parallel]
# -- Touch detection of putative synapses
#
# snudda prune <networkPath> [--mergeonly] [--parallel]
# -- Prune the synapses
#
# snudda input <networkPath> [--input yourInputConfig] [--parallel]
#
# snudda export <networkPath>
# -- Export to SONATA format (optional)
#
# snudda simulate <networkPath>
#
# snudda analyse <networkPath>
#
#
# snudda help me

# Johannes Hjorth, Royal Institute of Technology (KTH)
# Human Brain Project 2019

#
# This open source software code was developed in part or in whole in
# the Human Brain Project, funded from the European Union's Horizon
# 2020 Framework Programme for Research and Innovation under Specific
# Grant Agreements No. 720270 and No. 785907 (Human Brain Project SGA1
# and SGA2).
#


import os
import sys
import timeit
from collections import OrderedDict

# import pkg_resources
from importlib import resources

import json
import numpy as np

from snudda.utils import snudda_path
from snudda.utils.snudda_path import snudda_isfile, get_snudda_data


def get_data_file(*dirs):
    path = os.path.join("data", *dirs)

    try:
        return resources.files(__package__).joinpath(path).as_posix()
    except FileNotFoundError:
        raise FileNotFoundError(f"Data file '{path}' not found")


class Snudda(object):

    """ Wrapper class, calls Snudda helper functions """

    def __init__(self, network_path, parallel=False, ipython_profile=None, rc=None):

        """
        Instantiates Snudda
        :param network_path: Location of Snudda network
        """

        self.network_path = network_path

        self.d_view = None
        self.rc = rc
        self.slurm_id = 0
        self.logfile = None

        self.parallel = parallel
        self.ipython_profile = ipython_profile

        # Add current dir to python path
        sys.path.append(os.getcwd())

        self.start = timeit.default_timer()

    ############################################################################

    @staticmethod
    def help_info(args):
        """ Prints Snudda help """
        from snudda.help import snudda_help_text

    ############################################################################

    def init_config_wrapper(self, args):

        """
        Creates network-config.json in network_path.

        Args:
            args : command line arguments from argparse

        Example:
             snudda init -size 100 [-overwrite] [-randomseed 1234] [--profile] [--verbose] path
        """

        assert args.size is not None, "You need to specify --size when initialising config for the network"

        self.init_config(network_size=args.size,
                         snudda_data=args.snudda_data,
                         neurons_dir=args.neurons_dir,
                         connection_file=args.connection_file,
                         overwrite=args.overwrite,
                         random_seed=args.randomseed,
                         honor_stay_inside=args.stay_inside)

    def init_config(self,
                    network_size=None,
                    snudda_data=None,
                    struct_def=None,
                    neurons_dir=None,
                    connection_file=None,
                    honor_stay_inside=True,   # currently the cli.py defaults to sending False
                    overwrite=False,
                    random_seed=None):

        print(f"Legacy config creation.")

        # self.networkPath = args.path
        print("Creating config file")
        print(f"Network path: {self.network_path}")

        from snudda.init.init import SnuddaInit

        if struct_def is None:
            struct_def = {"Striatum": network_size,
                          "GPe": 0,
                          "GPi": 0,
                          "SNr": 0,
                          "STN": 0,
                          "Cortex": 0,  # Cortex and thalamus axons disabled right now, set to N > 0 to include a few
                          "Thalamus": 0}

        if not overwrite:
            if os.path.exists(self.network_path):
                print(f"Network path {self.network_path} already exists (aborting to prevent accidental overwriting)."
                      "\nCall snudda init with --overwrite to override and overwrite the old data.")
                sys.exit(1)

        self.make_dir_if_needed(self.network_path)

        config_file = os.path.join(self.network_path, "network-config.json")
        SnuddaInit(network_path=self.network_path,
                   struct_def=struct_def,
                   neurons_dir=neurons_dir,
                   snudda_data=snudda_data,
                   config_file=config_file,
                   honor_stay_inside=honor_stay_inside,
                   random_seed=random_seed,
                   connection_override_file=connection_file)

        if network_size is not None and network_size > 1e5:
            print(f"Make sure there is enough disk space in {self.network_path}")
            print("Large networks take up ALOT of space")

    ############################################################################

    def init_tiny(self, neuron_paths, neuron_names, number_of_neurons,
                  snudda_data=None,
                  morphology_key=None, parameter_key=None,
                  connection_config=None, random_seed=None,
                  density=80500, d_min=15e-6,
                  volume_mesh="cube",
                  side_len=None,
                  slice_depth=None):

        """
            network_path : Network path

        """

        from snudda.init.init import SnuddaInit
        from snudda.place import create_cube_mesh

        n_total = np.sum(number_of_neurons)

        if snudda_data is None:
            print(f"Warning, no snudda_data set!")

        si = SnuddaInit(network_path=self.network_path,
                        snudda_data=snudda_data,
                        random_seed=random_seed)

        if volume_mesh == "cube":
            volume_name = "Cube"

            if side_len is None:
                if n_total > 1:
                    side_len = (n_total/density)**(1/3)*1e-3
                else:
                    # When placing one neuron in a really small volume, sometimes it is too small
                    side_len = (5/density)**(1/3)*1e-3

            si.define_structure(struct_name="Cube",
                                struct_mesh="cube",
                                d_min=d_min,
                                struct_centre=(0.0, 0.0, 0.0),
                                side_len=side_len,
                                num_neurons=n_total,
                                n_putative_points=n_total*5)

        elif volume_mesh == "slice":
            volume_name = "Slice"

            si.define_structure(struct_name="Slice",
                                struct_mesh="slice",
                                d_min=d_min,
                                side_len=side_len,
                                slice_depth=slice_depth,
                                struct_centre=(0.0, 0.0, 0.0),
                                num_neurons=n_total,
                                n_putative_points=n_total*5)
        else:
            volume_name = "Custom"

            si.define_structure(struct_name=volume_name,
                                struct_mesh=volume_mesh,
                                d_min=d_min,
                                num_neurons=n_total,
                                n_putative_points=n_total*5)

        if connection_config is not None:
            if isinstance(connection_config, dict):
                si.replace_connectivity(connection_dict=connection_config)
            else:
                si.replace_connectivity(connection_file=connection_config)

        if isinstance(neuron_paths, str):
            neuron_paths = [neuron_paths]

        if isinstance(neuron_names, str):
            neuron_names = [neuron_names]

        if isinstance(number_of_neurons, int):
            number_of_neurons = [int(number_of_neurons / len(neuron_paths)) for x in neuron_names]

        if isinstance(morphology_key, str):
            morphology_key = [morphology_key]

        if isinstance(parameter_key, str):
            parameter_key = [parameter_key]

        assert (morphology_key is None and parameter_key is None) or \
            len(morphology_key) == len(parameter_key) == len(neuron_paths)

        for idx, (path, name, cnt) in enumerate(zip(neuron_paths, neuron_names, number_of_neurons)):
            if name in si.network_data["regions"][volume_name]["neurons"]:
                raise ValueError(f"neuron name {name} defined more than once")

            si.add_neurons(name=name, neuron_dir=path, region_name=volume_name, num_neurons=cnt)

            if morphology_key is not None:
                si.network_data["regions"][volume_name]["neurons"][name]["parameter_key"] = parameter_key[idx]
                si.network_data["regions"][volume_name]["neurons"][name]["morphology_key"] = morphology_key[idx]

        si.write_json()

        return si

    ############################################################################

    def import_config_wrapper(self, args):

        self.import_config(network_config_file=args.config_file,
                           snudda_data=args.snudda_data,
                           overwrite=args.overwrite)

    def import_config(self, network_config_file, snudda_data=None, overwrite=False):

        from snudda.init.init_config import ConfigParser

        conf = ConfigParser(config_file=network_config_file, snudda_data=snudda_data)
        conf.parse_config()

        if not os.path.isdir(self.network_path):
            print(f"Creating directory {self.network_path}")
            os.makedirs(self.network_path)

        new_config_file = os.path.join(self.network_path, "network-config.json")

        if os.path.isfile(new_config_file) and not overwrite:
            print(f"\n!!! ERROR: File already exists: {new_config_file}\nSet 'overwrite' to overwrite old file.\n")
            exit(-1)

        conf.replace_network_path(network_path=os.path.abspath(self.network_path))
        conf.write_config(new_config_file)

    ############################################################################

    def create_network(self, honor_morphology_stay_inside=True, verbose=False, keep_rc_active=False,
                       input_config=None, use_meta_input=True):

        # This function uses one remote client, that is used by all processes

        # This is a helper function, to create the full network
        self.place_neurons(honor_morphology_stay_inside=honor_morphology_stay_inside, verbose=verbose, keep_rc_active=True)
        self.detect_synapses(verbose=verbose, keep_rc_active=True)
        self.prune_synapses(verbose=verbose, keep_rc_active=keep_rc_active or input_config is not None)

        if input_config is not None:
            self.setup_input(input_config=input_config, keep_rc_active=keep_rc_active, use_meta_input=use_meta_input)

    def create_network_wrapper(self, args):

        self.create_network(input_config=args.input, verbose=args.verbose)

    ############################################################################

    def place_neurons_wrapper(self, args):
        """
        Places neurons in 3D space. Creates network-neuron-positions.hdf5 in network_path.

        Args:
            args : command line arguments from argparse

        Example:
            snudda place [--profile] [--verbose] [--h5legacy] [-parallel] path

        """

        if args.h5legacy:
            h5libver = "earliest"
        else:
            h5libver = "latest"  # default

        self.place_neurons(random_seed=args.randomseed,
                           parallel=args.parallel,
                           ipython_profile=args.ipython_profile,
                           ipython_timeout=args.ipython_timeout,
                           h5libver=h5libver,
                           verbose=args.verbose,
                           honor_morphology_stay_inside=args.stay_inside)

    def place_neurons(self,
                      random_seed=None,
                      parallel=None,
                      ipython_profile=None,
                      ipython_timeout=120,
                      h5libver="latest",
                      verbose=False,
                      honor_morphology_stay_inside=True,
                      keep_rc_active=False):

        if parallel is None:
            parallel = self.parallel

        if ipython_profile is None:
            ipython_profile = self.ipython_profile

        # self.networkPath = args.path
        print("Placing neurons")
        print(f"Network path: {self.network_path}")

        log_file_name = os.path.join(self.network_path, "log", "place-neurons.txt")

        self.setup_log_file(log_file_name)  # sets self.logFile

        if parallel:
            self.setup_parallel(ipython_profile=ipython_profile, timeout=ipython_timeout)  # sets self.d_view

        from snudda.place.place import SnuddaPlace

        sp = SnuddaPlace(network_path=self.network_path,
                         log_file=self.logfile,
                         verbose=verbose,
                         d_view=self.d_view,
                         h5libver=h5libver,
                         random_seed=random_seed,
                         morphologies_stay_inside=honor_morphology_stay_inside)

        sp.place()
        sp.close_log_file()

        self.cleanup_workers()

        if not keep_rc_active:
            self.close_rc()

        return sp

    ############################################################################

    def detect_synapses_wrapper(self, args):
        """
        Synapse touch detection. Writes results to network_path/voxels (one file per hypervoxel).
        Also adds synapse projections between structures.
        Results written to network-projection-synapses.hdf5 in network_path.

        Args:
            args : command line arguments from argparse

        Example:
            snudda detect [-cont] [-hvsize HVSIZE] [--volumeID VOLUMEID] [--profile] [--verbose] [--h5legacy] [-parallel] path

        """

        if args.hvsize is not None:
            hyper_voxel_size = int(args.hvsize)
        else:
            hyper_voxel_size = 100

        if args.h5legacy:
            h5libver = "earliest"
        else:
            h5libver = "latest"  # default

        self.detect_synapses(random_seed=args.randomseed,
                             parallel=args.parallel,
                             ipython_profile=args.ipython_profile,
                             ipython_timeout=args.ipython_timeout,
                             hyper_voxel_size=hyper_voxel_size,
                             volume_id=args.volume_id,
                             h5libver=h5libver,
                             verbose=args.verbose,
                             cont=args.cont)

    def detect_synapses(self,
                        random_seed=None,
                        parallel=None,
                        ipython_profile=None,
                        ipython_timeout=120,
                        hyper_voxel_size=100,
                        volume_id=None,
                        h5libver="latest",
                        verbose=False,
                        cont=False,
                        keep_rc_active=False):

        if parallel is None:
            parallel = self.parallel

        if ipython_profile is None:
            ipython_profile = self.ipython_profile

        # self.networkPath = args.path
        print("Touch detection")
        print(f"Network path: {self.network_path}")

        log_dir = os.path.join(self.network_path, "log")
        if not os.path.exists(log_dir):
            print(f"Creating directory {log_dir}")
            os.makedirs(log_dir, exist_ok=True)

        config_file = os.path.join(self.network_path, "network-config.json")
        position_file = os.path.join(self.network_path, "network-neuron-positions.hdf5")
        log_filename = os.path.join(self.network_path, "log", "touch-detection.txt")
        save_file = os.path.join(self.network_path, "voxels", "network-putative-synapses.hdf5")

        voxel_dir = os.path.join(self.network_path, "voxels")
        self.make_dir_if_needed(voxel_dir)

        self.setup_log_file(log_filename)  # sets self.logfile

        if parallel:
            self.setup_parallel(ipython_profile=ipython_profile, timeout=ipython_timeout)  # sets self.d_view

        from snudda.detect.detect import SnuddaDetect

        # You can now setup SnuddaDetect with only network_path and it will use default values
        # for config_file, position_file, logfile, save_file
        sd = SnuddaDetect(config_file=config_file,
                          position_file=position_file,
                          logfile=self.logfile,
                          save_file=save_file,
                          slurm_id=self.slurm_id,
                          volume_id=volume_id,
                          rc=self.rc,
                          hyper_voxel_size=hyper_voxel_size,
                          h5libver=h5libver,
                          random_seed=random_seed,
                          verbose=verbose)

        if cont:
            # Continue previous run
            print("Continuing previous touch detection")
            sd.detect(restart_detection_flag=False)
        else:
            sd.detect(restart_detection_flag=True)

        sd.close_log_file()

        # Also run SnuddaProject to handle projections between volume

        from snudda.detect.project import SnuddaProject

        sp = SnuddaProject(network_path=self.network_path)
        sp.project()

        self.cleanup_workers()

        if not keep_rc_active:
            self.close_rc()

        return sd, sp

    ############################################################################

    def prune_synapses_wrapper(self, args):

        """
        Merges data from synapse detection, then does synapse pruning. Writes results to network-synapses.hdf5 in network_path.

        Args:
            args : command line arguments from argparse

        Example:
            snudda prune [--configFile CONFIG_FILE] [--profile] [--verbose] [--h5legacy] [--keepfiles] [-parallel] path
        """

        if args.h5legacy:
            h5libver = "earliest"
        else:
            h5libver = "latest"  # default

        self.prune_synapses(config_file=args.config_file,
                            random_seed=args.randomseed,
                            parallel=args.parallel,
                            ipython_profile=args.ipython_profile,
                            ipython_timeout=args.ipython_timeout,
                            verbose=args.verbose,
                            keep_files=args.keepfiles,
                            save_putative_synapses = args.save_putative)

    def prune_synapses(self,
                       config_file=None,
                       random_seed=None,
                       parallel=None,
                       ipython_profile=None,
                       ipython_timeout=120,
                       h5libver="latest",
                       verbose=False,
                       keep_files=False,
                       save_putative_synapses=False,
                       keep_rc_active=False):

        if parallel is None:
            parallel = self.parallel

        if ipython_profile is None:
            ipython_profile = self.ipython_profile

        # self.networkPath = args.path
        print("Prune synapses")
        print(f"Network path: {self.network_path}")

        from snudda.detect.prune import SnuddaPrune

        log_filename = os.path.join(self.network_path, "log", "synapse-pruning.txt")

        self.setup_log_file(log_filename)  # sets self.logfile

        if parallel:
            self.setup_parallel(ipython_profile=ipython_profile, timeout=ipython_timeout)  # sets self.d_view

        # Optionally set this
        scratch_path = None

        sp = SnuddaPrune(network_path=self.network_path,
                         logfile=self.logfile,
                         logfile_name=log_filename,
                         config_file=config_file,
                         d_view=self.d_view,
                         scratch_path=scratch_path,
                         h5libver=h5libver,
                         random_seed=random_seed,
                         verbose=verbose,
                         keep_files=keep_files or save_putative_synapses)

        sp.prune()

        if save_putative_synapses:
            sp.save_putative_synapses()

        sp.close_log_file()

        self.cleanup_workers()

        if not keep_rc_active:
            self.close_rc()

        return sp

    ############################################################################

    def setup_input_wrapper(self, args):
        """
        Creates synaptic input for network based on input.json, writes input-spikes.hdf5 in network_path.
        Args:
            args : command line arguments from argparse

        Example:
            snudda input [--input INPUT] [--inputFile INPUT_FILE] [--networkFile NETWORK_FILE] [--time TIME] [-randomseed 123] [--profile] [--verbose] [--h5legacy] [-parallel] path
        """

        if args.h5legacy:
            h5libver = "earliest"
        else:
            h5libver = "latest"  # default

        self.setup_input(network_file=args.network_file,
                         input_file=args.input_file,
                         input_config=args.input,
                         input_time=args.time,
                         random_seed=args.randomseed,
                         use_meta_input=not args.no_meta_input,
                         h5libver=h5libver,
                         parallel=args.parallel,
                         ipython_profile=args.ipython_profile,
                         ipython_timeout=args.ipython_timeout,
                         verbose=args.verbose)

    def setup_input(self,
                    network_file=None,
                    input_file=None,
                    input_config=None,
                    input_time=None,
                    use_meta_input=True,
                    random_seed=None,
                    h5libver="latest",
                    parallel=None,
                    ipython_profile=None,
                    ipython_timeout=120,
                    verbose=False,
                    keep_rc_active=False):

        if parallel is None:
            parallel = self.parallel

        if ipython_profile is None:
            ipython_profile = self.ipython_profile

        print("Setting up inputs, assuming input.json exists")
        log_filename = os.path.join(self.network_path, "log", "setup-input.txt")
        self.setup_log_file(log_filename)  # sets self.logfile

        if parallel:
            self.setup_parallel(ipython_profile=ipython_profile, timeout=ipython_timeout)  # sets self.d_view

        from snudda.input.input import SnuddaInput

        if input_config is None:
            input_config = os.path.join(self.network_path, "input.json")

        snudda_data = get_snudda_data(network_path=self.network_path)
        if not snudda_isfile(input_config, snudda_data=snudda_data):
            print(f"Missing input config file: {input_config}")
            return

        if network_file is None:
            network_file = os.path.join(self.network_path, "network-synapses.hdf5")

        if input_file is None:
            input_file = os.path.join(self.network_path, "input-spikes.hdf5")

        print(f"Writing input spikes to {input_file}")

        si = SnuddaInput(input_config_file=input_config,
                         hdf5_network_file=network_file,
                         spike_data_filename=input_file,
                         time=input_time,
                         logfile=self.logfile,
                         rc=self.rc,
                         random_seed=random_seed,
                         h5libver=h5libver,
                         verbose=verbose,
                         use_meta_input=use_meta_input)
        si.generate()
        si.close_log_file()

        self.cleanup_workers()

        if not keep_rc_active:
            self.close_rc()

        return si

    ############################################################################

    def export_to_SONATA_wrapper(self, args):
        """
        Export network to SONATA files. (Currently not functioning)

        Args:
            args : command line arguments from argparse
        """

        self.export_to_SONATA(network_file=args.network_file,
                              input_file=args.input_file)

    def export_to_SONATA(self, network_file=None, input_file=None):

        from snudda.utils.export_sonata import ExportSonata

        print("Exporting to SONATA format")
        print(f"Network path: {self.network_path}")

        if network_file is None:
            network_file = os.path.join(self.network_path, "network-synapses.hdf5")

        if input_file is None:
            input_file = os.path.join(self.network_path, "input-spikes.hdf5")

        out_dir = os.path.join(self.network_path, "SONATA")

        cn = ExportSonata(network_file=network_file, input_file=input_file, out_dir=out_dir)

    ############################################################################

    @staticmethod
    def compile_mechanisms(mech_dir=None, snudda_data=None):

        if not mech_dir:
            mech_dir = os.path.realpath(snudda_path.snudda_parse_path(os.path.join("$DATA", "neurons", "mechanisms"),
                                                                      snudda_data=snudda_data))

        if not os.path.exists("x86_64") and not os.path.exists("nrnmech.dll")\
                and not os.path.exists("aarch64") and not os.path.exists("arm64"):

            from mpi4py import MPI  # This must be imported before neuron, to run parallel
            from neuron import h
            pc = h.ParallelContext()

            if pc.id() == 0:
                # Only run this on master node
                print(f"Running on master node:  nrnivmodl {mech_dir}")
                os.system(f"nrnivmodl {mech_dir}")
            else:
                print("Worker waiting for master node to compile NEURON modules.")

            pc.barrier()

            try:
                if os.path.exists("nrnmech.dll"):
                    mech_path = "nrnmech.dll"
                elif os.path.exists("x86_64"):
                    mech_path = "x86_64/.libs/libnrnmech.so"
                elif os.path.exists("aarch64"):
                    mech_path = "aarch64/.libs/libnrnmech.so"
                elif os.path.exists("arm64"):
                    mech_path = "arm64/.libs/libnrnmech.so"
                else:
                    print(f"Could not find compiled mechanisms. Compile using 'nrnivmodl {mech_dir}' "
                          f"and retry simulation.")
                    sys.exit(-1)

                print(f"Loading mechanisms from '{os.path.abspath(mech_path)}'")
                h.nrn_load_dll(mech_path)

            except:
                import traceback
                print(f"Error while loading mechanisms:\n{traceback.format_exc()}")


        else:
            print("NEURON mechanisms already compiled, make sure you have the correct version of NEURON modules."
                  "\nIf you delete x86_64, aarch64, arm64 directories (or nrnmech.dll) "
                  "then you will force a recompilation of the modules.")

    ############################################################################

    def simulate_wrapper(self, args):
        """
        Simulate network. Writes results to network_path/simulation.

        Args:
            args : command line arguments from argparse

        Example:
            snudda simulate [--networkFile NETWORK_FILE] [--inputFile INPUT_FILE] [--time TIME]
            [--spikesOut SPIKES_OUT] [--noVolt] [--disableGJ]
            [-mechdir MECH_DIR] [--profile] [--verbose] [--exportCoreNeuron] path
        """

        print(f"args: {args}")

        assert args.enable_rxd_neuromodulation is None or args.disable_rxd_neuromodulation is None, \
            "You can only specify enable_rxd_neuromodulation or disable_rxd_neuromodulation, not both"

        use_rxd_neuromodulation = None

        if args.enable_rxd_neuromodulation == True:
            use_rxd_neuromodulation = True

        if args.disable_rxd_neuromodulation == True:
            use_rxd_neuromodulation = False

        sim = self.simulate(network_file=args.network_file, input_file=args.input_file,
                            output_file=args.output_file, snudda_data=args.snudda_data,
                            time=args.time,
                            mech_dir=args.mech_dir,
                            # neuromodulation=args.neuromodulation,
                            disable_synapses=args.disable_synapses,
                            disable_gj=args.disable_gj,
                            record_volt=args.record_volt,
                            record_all=args.record_all,
                            simulation_config=args.simulation_config,
                            export_core_neuron=args.export_core_neuron,
                            use_rxd_neuromodulation=use_rxd_neuromodulation,
                            verbose=args.verbose)

        sim.clear_neuron()

    def simulate(self,
                 network_file=None,
                 input_file=None,
                 output_file=None,
                 snudda_data=None,
                 time=None,
                 mech_dir=None,
                 # neuromodulation=None,
                 disable_synapses=None,
                 disable_gj=False,
                 record_volt=True,
                 record_all=False,
                 sample_dt=None,
                 simulation_config=None,
                 export_core_neuron=False,
                 use_rxd_neuromodulation=None,
                 verbose=False):

        start = timeit.default_timer()

        from mpi4py import MPI  # This must be imported before neuron, to run parallel

        # Initialize MPI
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size()
        print(f"MPI Rank: {rank}, Size: {size}")

        from neuron import h
        pc = h.ParallelContext()

        if network_file is None:
            network_file = os.path.join(self.network_path, "network-synapses.hdf5")

        if input_file is None:
            putative_input_file = os.path.join(self.network_path, "input-spikes.hdf5")
            if os.path.isfile(putative_input_file):
                input_file = putative_input_file

        if output_file is None:
            output_file = os.path.join(self.network_path, "simulation", "output.hdf5")

        self.make_dir_if_needed(os.path.join(self.network_path, "simulation"))

        print(f"Using input file {input_file}")

        # nWorkers = args.ncores
        # print("Using " + str(nWorkers) + " workers for neuron")

        # Problems with nested symbolic links when the second one is a relative
        # path going beyond the original base path

        if mech_dir is None:
            # Take into account which SNUDDA_DATA the user wants to use
            from snudda.utils.snudda_path import get_snudda_data
            snudda_data = get_snudda_data(snudda_data=snudda_data,
                                          network_path=self.network_path)

            mech_dir = os.path.realpath(snudda_path.snudda_parse_path(os.path.join("$DATA", "neurons", "mechanisms"),
                                                                      snudda_data))

        self.compile_mechanisms(mech_dir=mech_dir)

        save_dir = os.path.join(os.path.dirname(network_file), "simulation")

        if not os.path.exists(save_dir):
            print(f"Creating directory {save_dir}")
            os.makedirs(save_dir, exist_ok=True)

        # Get the SlurmID, used in default file names
        slurm_id = os.getenv('SLURM_JOBID')

        if slurm_id is None:
            slurm_id = str(666)

        if disable_gj:
            print("!!! WE HAVE DISABLED GAP JUNCTIONS !!!")

        if disable_synapses:
            print("!!! SYNAPSES DISABLED")

        log_file = os.path.join(os.path.dirname(network_file), "log", "network-simulation-log.txt")

        log_dir = os.path.join(os.path.dirname(network_file), "log")
        if not os.path.exists(log_dir):
            print(f"Creating directory {log_dir}")
            os.makedirs(log_dir, exist_ok=True)

        from snudda.simulate.simulate import SnuddaSimulate

        # Simulate is deterministic, no random seed.
        sim = SnuddaSimulate(network_file=network_file,
                             input_file=input_file,
                             output_file=output_file,
                             snudda_data=snudda_data,
                             disable_gap_junctions=disable_gj,
                             disable_synapses=disable_synapses,
                             log_file=log_file,
                             simulation_config=simulation_config,
                             sample_dt=sample_dt,
                             use_rxd_neuromodulation=use_rxd_neuromodulation,
                             verbose=verbose)
        sim.setup()
        # sim.add_external_input() # -- now done in setup

        sim.check_memory_status()

        if record_volt:
            # sim.add_volt_recording_all()
            # Either use record_volt or specify "record_all_soma" in simulation_config file
            sim.add_volt_recording_soma()
            # sim.addRecordingOfType("dSPN",5) # Side len let you record from a subset

        if record_all:
            record_cell_id = np.array([int(x) for x in record_all.split(",")])
            sim.add_volt_recording_all(cell_id=record_cell_id)
            sim.add_synapse_current_recording_all(record_cell_id)

        if time is not None:
            t_sim = time * 1000  # Convert from s to ms for Neuron simulator
        else:
            # Will attempt to read "time" from simulation_config
            t_sim = None

        if export_core_neuron:
            sim.export_to_core_neuron()
            return  # We do not run simulation when exporting to core neuron

        sim.check_memory_status()
        if t_sim is None or t_sim > 0:
            print(f"Running simulation for {t_sim} ms.")
            sim.run(t_sim)  # In milliseconds

            print("Simulation done, saving output")
            sim.write_output()
        else:
            print(f"Time set to {t_sim} ms. No simulation run.")

        stop = timeit.default_timer()
        if sim.pc.id() == 0:
            print(f"Program run time: {stop - start:.1f}s")

        # OBS! You want to do sim.clear_neuron() after the simulation if you need
        #      to setup another neuron simulation afterwards.

        # sim.plot()
        return sim

    ############################################################################

    def analyse(self, args):

        print("Add analysis code here, see Network_analyse.py")

    ############################################################################

    def setup_parallel(self, ipython_profile=None, timeout=120, n_workers=None):
        """Setup ipyparallel workers."""

        if self.rc is not None:
            self.logfile.write("setup_parallel: ipyparallel clients already setup, ignoring. ")
            return

        self.slurm_id = os.getenv('SLURM_JOBID')

        if self.slurm_id is None:
            self.slurm_id = 0
        else:
            self.slurm_id = int(self.slurm_id)

        self.logfile.write(f"Using slurm_id: {self.slurm_id}")

        if ipython_profile is None:
            ipython_profile = os.getenv('IPYTHON_PROFILE')

        if not ipython_profile:
            ipython_profile = "default"

        ipython_dir = os.getenv('IPYTHONDIR')
        if not ipython_dir:
            ipython_dir = os.path.join(os.path.abspath(os.getcwd()), ".ipython")

        self.logfile.write('Creating ipyparallel client\n')

        from ipyparallel import Client

        u_file = os.path.join(ipython_dir, f"profile_{ipython_profile}", "security", "ipcontroller-client.json")
        print(f"Reading IPYPARALLEL connection info from {u_file}\n")
        self.logfile.write(f"Reading IPYPARALLEL connection info from {u_file}\n")
        # self.rc = Client(profile=ipython_profile, url_file=u_file, timeout=120, debug=False)
        self.rc = Client(profile=ipython_profile, connection_info=u_file, timeout=timeout, debug=False)

        # Detect expected number of workers
        if n_workers is None:
            n_workers = self.get_expected_engines(ipython_profile=ipython_profile)
        else:
            print(f"Expecting user defined {n_workers =}")

        # Wait for engines to register
        try:
            self.rc.wait_for_engines(n=n_workers, timeout=timeout)
        except Exception as e:
            raise RuntimeError(f"Engines did not start within {timeout} seconds: {e}")

        if not self.rc.ids:
            raise RuntimeError("No engines registered. Ensure engines are started before proceeding.")

        self.logfile.write(f'Client IDs: {self.rc.ids}')

        # http://davidmasad.com/blog/simulation-with-ipyparallel/
        # http://people.duke.edu/~ccc14/sta-663-2016/19C_IPyParallel.html
        self.d_view = self.rc.direct_view(targets='all')  # rc[:] # Direct view into clients

        # Make sure SNUDDA_DATA is set on the workers, this might be needed if ipcluster
        # is started before SNUDDA_DATA is set
        if os.getenv('SNUDDA_DATA') is not None:
            print(f"Setting SNUDDA_DATA environment variable on workers to {os.getenv('SNUDDA_DATA')}")

            self.d_view.execute("import os")
            self.d_view.execute(f"os.environ['SNUDDA_DATA'] = '{os.getenv('SNUDDA_DATA')}'", block=True)

    def get_expected_engines(self, ipython_profile="default"):

        """Get the number of expected engines from the ipcontroller configuration."""
        ipython_dir = os.getenv('IPYTHONDIR', os.path.join(os.path.expanduser("~"), ".ipython"))
        engine_log_file = os.path.join(ipython_dir, f"profile_{ipython_profile}", "log", "engines.json")

        num_workers = int(os.getenv('SLURM_NTASKS', 1))  # Default to 1 if not in SLURM

        # Sometimes not all SLURM workers are used... check ipyparallel config file
        if os.path.exists(engine_log_file):
            with open(engine_log_file, "r") as f:
                log = json.load(f)
                num_engines = len(log["engines"])
                print(f"Detected {num_engines} engines from {engine_log_file}")

                if os.getenv('SLURM_NTASKS') is not None:
                    if num_workers != num_engines:
                        print(f"Warning, {engine_log_file} does not have SLURM_NTASKS ({num_workers}) engines (this could be intentional)")

                return num_engines
        else:
            print(f"No engine log file {engine_log_file}")

        print(f"Expecting {num_workers}, set in SLURM_NTASKS")

        return num_workers

    ############################################################################

    def cleanup_workers(self):

        if self.logfile is not None and not self.logfile.closed:
            self.logfile.write(f"Calling cleanup on workers.")
        else:
            print(f"Calling cleanup on workers (log file already closed).")
        # Cleanup, and do garbage collection

        # spd uses the sd log file
        close_logs_cmd = ("if 'sm' in locals() and sm is not None: sm.close_log_file()\n"
                          "if 'sd' in locals() and sd is not None: sd.close_log_file()\n"
                          "if 'sp' in locals() and sp is not None: sp.close_log_file()\n"
                          "if 'nl' in locals() and nl is not None: nl.close_log_file()\n")

        if self.d_view is not None:
            self.d_view.execute(close_logs_cmd, block=True)

        clean_cmd = ("sm = None\nsd = None\nsp = None\nspd = None\nnl = None\nsim = None"
                     "\ninner_mask = None\nmin_max = None\nneuron_hv_list = None"
                     "\nsyn_before = None\nsyn_after = None\ninpt = None"
                     "\nmerge_result_syn = None\nmerge_result_gj = None"
                     "\nimport gc\ngc.collect()")

        if self.d_view is not None:
            self.d_view.execute(clean_cmd, block=True)

    ############################################################################

    def close_rc(self):
        if self.rc is not None:
            print("Closing remote client (ipcluster kept running)")
            self.rc.close()
            self.rc = None

    def stop_parallel(self):

        if self.rc is not None:
            print("Stopping ipyparallel")
            self.rc.shutdown(hub=True)
            self.rc.close()
            self.rc = None

    ############################################################################

    def setup_log_file(self, log_file_name):
        """
        Open log files for writing.

        Args:
            log_file_name (str) : Path to log file
        """

        data_dir = os.path.dirname(log_file_name)

        self.make_dir_if_needed(data_dir)

        try:
            self.logfile = open(log_file_name, 'w')
            self.logfile.write('Starting log file\n')
        except:
            print("Unable to set up log file " + str(log_file_name))
            self.logfile = None

    ############################################################################

    def close_log_file(self):
        """ Close log file. """

        stop = timeit.default_timer()

        print(f"\nExecution time: {stop - self.start:.1f}s")

        if self.logfile is not None and not self.logfile.closed:
            self.logfile.write(f"Execution time: {stop - self.start:.1f}s")
            self.logfile.write("End of log. Closing file.")
            self.logfile.close()
            self.logfile = None

    ############################################################################

    @staticmethod
    def make_dir_if_needed(dir_path):
        """ Creates directory if missing. """

        if not os.path.exists(dir_path):
            print("Creating missing directory " + dir_path)
            try:
                os.makedirs(dir_path)
                print("Created directory " + dir_path)
            except:
                print("Failed to create dir " + dir_path)

    ############################################################################

    @staticmethod
    def cleanup_workers_rc(rc):

        if rc is None:
            return

        d_view = rc.direct_view(targets='all')  # rc[:] # Direct view into clients

        clean_cmd = ("sm = None\nsd = None\nsp = None\nspd = None\nnl = None\nsim = None"
                     "\ninner_mask = None\nmin_max = None\nneuron_hv_list = None"
                     "\nsyn_before = None\nsyn_after = None\ninpt = None"
                     "\nmerge_result_syn = None\nmerge_result_gj = None"
                     "\nimport gc\ngc.collect()")

        d_view.execute(clean_cmd, block=True)


##############################################################################


if __name__ == "__main__":

    # This is fix to handle if user calles python from within neuron
    import sys

    if '-python' in sys.argv:
        print("Network_simulate.py called through nrniv, fixing arguments")
        pythonidx = sys.argv.index('-python')
        if len(sys.argv) > pythonidx:
            sys.argv = sys.argv[pythonidx + 1:]

    from .cli import snudda_cli

    snudda_cli()



import os
import sys
import unittest
import numpy as np

from snudda.detect.detect import SnuddaDetect
from snudda.place.create_cube_mesh import create_cube_mesh
from snudda.place.place import SnuddaPlace

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))


class TestDetect(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_detect")

        create_cube_mesh(file_name=os.path.join(network_path, "mesh", "simple_mesh.obj"),
                         centre_point=(0, 0, 0),
                         side_len=500e-6)

        config_file = os.path.join(network_path, "network-config.json")
        position_file = os.path.join(network_path, "network-neuron-positions.hdf5")
        save_file = os.path.join(network_path, "voxels", "network-putative-synapses.hdf5")

        #  TODO: If d_view is None code run sin serial, add test parallel
        sp = SnuddaPlace(config_file=config_file, d_view=None, verbose=True)
        sp.place()

        # We want to load in the ball and stick neuron that has 20 micrometer soma diameter, and axon (along y-axis),
        # and dendrite along (x-axis) out to 100 micrometer distance from centre of soma.

        self.sd = SnuddaDetect(config_file=config_file, position_file=position_file,
                               save_file=save_file, rc=None,
                               hyper_voxel_size=130, verbose=True)

    def test_detect(self):

        neuron_positions = np.array([[0, 20, 0],  # Postsynaptiska
                                     [0, 40, 0],
                                     [0, 60, 0],
                                     [0, 80, 0],
                                     [0, 100, 0],
                                     [0, 120, 0],
                                     [0, 140, 0],
                                     [0, 160, 0],
                                     [0, 180, 0],
                                     [0, 200, 0],
                                     [20, 0, 0],  # Presynaptiska
                                     [40, 0, 0],
                                     [60, 0, 0],
                                     [80, 0, 0],
                                     [100, 0, 0],
                                     [120, 0, 0],
                                     [140, 0, 0],
                                     [160, 0, 0],
                                     [180, 0, 0],
                                     [200, 0, 0],
                                     [100, 100, -40],  # To get a gap junction
                                     ]) * 1e-6

        for idx, pos in enumerate(neuron_positions):
            self.sd.neurons[idx]["position"] = pos

        ang = -np.pi / 2
        R_x = np.array([[1, 0, 0],
                        [0, np.cos(ang), -np.sin(ang)],
                        [0, np.sin(ang), np.cos(ang)]])

        ang = np.pi / 2
        R_y = np.array([[np.cos(ang), 0, np.sin(ang)],
                        [0, 1, 0],
                        [-np.sin(ang), 0, np.cos(ang)]])

        ang = -np.pi / 2
        R_gj = np.array([[np.cos(ang), 0, np.sin(ang)],
                         [0, 1, 0],
                         [-np.sin(ang), 0, np.cos(ang)]])

        for idx in range(0, 10):  # Post synaptic neurons
            self.sd.neurons[idx]["rotation"] = R_x

        for idx in range(10, 20):  # Presynaptic neurons
            self.sd.neurons[idx]["rotation"] = R_y

        self.sd.neurons[20]["rotation"] = R_gj

        self.sd.detect(restart_detection_flag=True)

        synapse_voxel_loc = self.sd.hyper_voxel_synapses[:self.sd.hyper_voxel_synapse_ctr, 2:5]
        synapse_coords = synapse_voxel_loc * self.sd.voxel_size + self.sd.hyper_voxel_origo

        fig_path = os.path.join("networks", "network_testing_detect", "figures")

        if not os.path.exists(fig_path):
            os.mkdir(fig_path)

        if False:  # Set to True to include plot
            self.sd.plot_hyper_voxel(plot_neurons=True, fig_file_name="touch-detection-validation")

        # TODO: Also add tests for soma-axon synapses

        # Check location and source, targets.
        with self.subTest(stage="gap_junction_check"):
            self.assertEqual(self.sd.hyper_voxel_gap_junction_ctr, 1)
            self.assertTrue((self.sd.hyper_voxel_gap_junctions[0, 0:2] == [4, 20]).all())
            self.assertTrue(self.compare_voxels_to_coordinates(self.sd.hyper_voxel_gap_junctions[0, 6:9],
                                                               np.array([100, 100, 0]) * 1e-6))
        with self.subTest(stage="synapses_check"):
            print(f"synapse ctr {self.sd.hyper_voxel_synapse_ctr}")

            self.assertEqual(self.sd.hyper_voxel_synapse_ctr, 101)

            for post_id in range(0, 10):
                for pre_id in range(10, 20):
                    self.assertEqual(self.check_neuron_pair_has_synapse(pre_id, post_id), 1)

            for post_id in range(0, 10):
                for pre_id in range(0, 10):
                    self.assertFalse(self.check_neuron_pair_has_synapse(pre_id, post_id))

            for post_id in range(10, 20):
                for pre_id in range(10, 20):
                    self.assertFalse(self.check_neuron_pair_has_synapse(pre_id, post_id))

        with self.subTest(stage="synapse_sorting_check"):
            syn = self.sd.hyper_voxel_synapses[:self.sd.hyper_voxel_synapse_ctr, :2]
            syn_order = syn[:, 1] * len(self.sd.neurons) + syn[:, 0]
            self.assertTrue((np.diff(syn_order) >= 0).all())

        with self.subTest(stage="gap_junction_sorting_check"):
            gj = self.sd.hyper_voxel_gap_junctions[:self.sd.hyper_voxel_gap_junction_ctr, :2]
            gj_order = gj[:, 1] * len(self.sd.neurons) + gj[:, 0]
            self.assertTrue((np.diff(gj_order) >= 0).all())

        with self.subTest(stage="synapse_conductance_check"):
            cond = self.sd.hyper_voxel_synapses[:self.sd.hyper_voxel_synapse_ctr, 11] * 1e-12
            self.assertTrue(0.8e-9 < np.mean(cond) < 1.4e-9)  # mean 1.1e-9
            self.assertTrue(0.5e-9 < np.std(cond) < 2.5e-9)  # std 1.5e-9

        with self.subTest(stage="gap_junction_conductance_check"):
            cond = self.sd.hyper_voxel_gap_junctions[:self.sd.hyper_voxel_gap_junction_ctr, 10] * 1e-12
            self.assertTrue(2e-10 < np.mean(cond) < 8e-10)  # mean 5e-10

            # Only one Gap junction, cant check std here
            # self.assertTrue(0.1e-10 < np.std(cond) < 2e-9)  # std 1e-10

        # We should probably store the matrix as unsigned.
        self.assertTrue((self.sd.hyper_voxel_synapses >= 0).all())
        self.assertTrue((self.sd.hyper_voxel_gap_junctions >= 0).all())

        with self.subTest(stage="resiz_matrix_check"):
            old = self.sd.hyper_voxel_synapses.copy()
            self.sd.resize_hyper_voxel_synapses_matrix()
            # Check all synapses are preserved when resizing
            self.assertTrue((self.sd.hyper_voxel_synapses[:old.shape[0], :] == old).all())
            # Check new rows are empty
            self.assertTrue((self.sd.hyper_voxel_synapses[old.shape[0]:, :] == 0).all())

        # These test drawing not essential to Snudda, quite slow.
        if False:
            with self.subTest(stage="export_voxel_vis"):
                self.sd.export_voxel_visualisation_csv(neuron_id=np.arange(0, 10))

            with self.subTest(stage="plot_hyper_voxel"):
                # Matplotlib is kind of slow
                self.sd.plot_neurons_in_hyper_voxel(neuron_id=np.arange(0, 10),
                                                    neuron_colour=np.zeros((10, 3)),
                                                    show_plot=False, dpi=90)

            with self.subTest(stage="example-draw"):
                # Just checking that the drawing works
                self.sd.test_voxel_draw()

        print("Checking detect done.")

    def check_neuron_pair_has_synapse(self, pre_neuron, post_neuron):

        connections = dict()

        for synapse_row in self.sd.hyper_voxel_synapses[0:self.sd.hyper_voxel_synapse_ctr, :]:

            loc = (synapse_row[0], synapse_row[1])
            if loc in connections:
                connections[loc] += 1
            else:
                connections[loc] = 1

        if (pre_neuron, post_neuron) in connections:
            # print(f"pre: {pre_neuron}, post: {post_neuron}, connections: {connections[(pre_neuron, post_neuron)]}")
            return connections[(pre_neuron, post_neuron)]
        else:
            # print(f"No connection between pre {pre_neuron} and post {post_neuron}")
            return False

    def convert_to_coordinates(self, voxel_xyz):
        return np.array(voxel_xyz) * self.sd.voxel_size + self.sd.hyper_voxel_origo

    def compare_voxels_to_coordinates(self, voxel_index, coordinates):
        return (np.abs(self.convert_to_coordinates(voxel_index) - np.array(coordinates)) < self.sd.voxel_size).all()

    def test_detect_lines(self):

        # Cases to test
        # id 0-9 connecting to id 10-19, connection angle is 0-45 degrees
        # id 20 is 4 micrometer and parallel to another dendrite, no intersection

        neuron_positions = np.array([[0, 0, 0],  # Postsynaptiska
                                     [10, 10, 15],
                                     [20, 20, 30],
                                     [30, 30, 45],
                                     [40, 40, 60],
                                     [50, 50, 75],
                                     [60, 60, 90],
                                     [70, 70, 105],
                                     [80, 80, 120],
                                     [90, 90, 135],
                                     [50, -100, 0],  # Presynaptiska
                                     [60, -90, 15],
                                     [70, -80, 30],
                                     [80, -70, 45],
                                     [90, -60, 60],
                                     [100, -50, 75],
                                     [110, -40, 90],
                                     [120, -30, 105],
                                     [130, -20, 120],
                                     [140, -10, 135],
                                     [230, 0, 4],  # 4 micrometers from first neuron
                                     ]) * 1e-6

        for idx, pos in enumerate(neuron_positions):
            self.sd.neurons[idx]["position"] = pos

        ang = -np.pi / 2
        R_x = np.array([[1, 0, 0],
                        [0, np.cos(ang), -np.sin(ang)],
                        [0, np.sin(ang), np.cos(ang)]])

        ang = np.pi / 2
        R_y = np.array([[np.cos(ang), 0, np.sin(ang)],
                        [0, 1, 0],
                        [-np.sin(ang), 0, np.cos(ang)]])

        for idx in range(0, 10):  # Post synaptic neurons
            self.sd.neurons[idx]["rotation"] = R_x

        for idx, ang in zip(range(10, 20), np.linspace(0, -np.pi / 4, 10)):  # Presynaptic neurons

            R_z = np.array([[np.cos(ang), -np.sin(ang), 0],
                            [np.sin(ang), np.cos(ang), 0],
                            [0, 0, 1]])

            print(f"idx = {idx}, ang = {ang}")

            self.sd.neurons[idx]["rotation"] = np.matmul(R_z, R_y)

        ang = np.pi / 2
        R_z = np.array([[np.cos(ang), -np.sin(ang), 0],
                        [np.sin(ang), np.cos(ang), 0],
                        [0, 0, 1]])

        self.sd.neurons[20]["rotation"] = np.matmul(R_z, R_y)

        self.sd.detect(restart_detection_flag=True)

        synapse_voxel_loc = self.sd.hyper_voxel_synapses[:self.sd.hyper_voxel_synapse_ctr, 2:5]
        synapse_coords = synapse_voxel_loc * self.sd.voxel_size + self.sd.hyper_voxel_origo

        if False:  # Set to True to include plot
            self.sd.plot_hyper_voxel(plot_neurons=True, fig_file_name="axon_dend_intersection_angle_0_45")

        with self.subTest(stage="synapses_check"):
            print(f"synapse ctr {self.sd.hyper_voxel_synapse_ctr}")

            # Depending on angle and location of voxels,
            # we can occasionally get more than one synapse at an intersection
            self.assertTrue(self.sd.hyper_voxel_synapse_ctr >= 10)

            # Verify that all pairs are connected
            for post_id in range(0, 10):
                pre_id = post_id + 10
                self.assertTrue(1 <= self.check_neuron_pair_has_synapse(pre_id, post_id) <= 2)

    # TODO: Gör ett test som distriburerar alla sections till sina respektive hypervoxlar,
    #       sen tvinga en touch detection som ignorerar den infon och tar med allt
    #       och se om några section id finns med som inte finns med i listorna...


if __name__ == '__main__':
    unittest.main()

# python3 -m unittest test_detect



import os.path
import unittest

import numpy as np
from snudda.detect import SnuddaDetect
from snudda.neurons.neuron_morphology_extended import NeuronMorphologyExtended
from snudda.place.rotation import SnuddaRotate

# We want to create a hyper voxel, then place simple ball and stick neurons to see that
# regardless of the direction they are rotated we get (approximately) the same number of
# voxels marked during touch detection


class TestVoxelCount(unittest.TestCase):

    def setUp(self) -> None:
        self.n_iter = 500

        self.sd = SnuddaDetect(hyper_voxel_size=160)  # Use 250 for real neurons
        self.sd.hyper_voxels = {0: {"random_seed": 1234}}
        self.sd.max_axon = 35  # 27 -- for real morphologies
        self.sd.max_dend = 30

        centre_pos = self.sd.num_bins * self.sd.voxel_size / 2

        self.sd.setup_hyper_voxel(hyper_voxel_id=0, hyper_voxel_origo=np.zeros((3,)))

        # neuron_file = "validation/shortballandstick/simpleshort.swc"
        # neuron_file = "validation/shortballandstick/simpleshortshort.swc"

        neuron_file = "validation/striatum-var/dspn/str-dspn-e150602_c1_D1-mWT-0728MSN01-v20211026/morphology/WT-0728MSN01-cor-rep-ax-res3-var8.swc"

        # neuron_file = "../../BasalGangliaData/data/neurons/striatum/dspn/str-dspn-e150917_c6_D1-m21-6-DE-v20211028/morphology/21-6-DE-cor-rep-ax-res3.swc"
        # neuron_file = "../../BasalGangliaData/data/neurons/striatum/dspn/str-dspn-e150602_c1_D1-mWT-0728MSN01-v20211026/morphology/WT-0728MSN01-cor-rep-ax-res3.swc"
        # neuron_file = "../../BasalGangliaData/data/neurons/striatum/dspn/str-dspn-e150602_c1_D1-mWT-0728MSN01-v20211026/morphology/WT-0728MSN01-cor-rep-ax-res3-var2.swc"

        self.neuron = NeuronMorphologyExtended(name="simple",
                                               swc_filename=neuron_file)

        rng = self.sd.hyper_voxel_rng
        neuron_pos = rng.uniform(low=centre_pos[0]-40e-6,
                                 high=centre_pos[0]+40e-6, size=(self.n_iter, 3))
        neuron_clones = [None]

        # Clone neuron and add multiple copies with different rotation centred in center
        for neuron_id in range(1, self.n_iter+1):
            print(f"Adding neuron {neuron_id}")

            if neuron_id == 1:
                clone_rotation = np.eye(3)
            else:
                clone_rotation = SnuddaRotate.rand_rotation_matrix(rng=rng)
            neuron_clones.append(self.neuron.clone(position=neuron_pos[neuron_id-1, :],
                                                   rotation=clone_rotation))

            self.sd.fill_voxels_dend(voxel_space=self.sd.dend_voxels,
                                     voxel_space_ctr=self.sd.dend_voxel_ctr,
                                     voxel_sec_id=self.sd.dend_sec_id,
                                     voxel_sec_x=self.sd.dend_sec_x,
                                     voxel_soma_dist=self.sd.dend_soma_dist,
                                     neuron=neuron_clones[neuron_id],
                                     neuron_id=neuron_id)

            self.sd.fill_voxels_axon(voxel_space=self.sd.axon_voxels,
                                     voxel_space_ctr=self.sd.axon_voxel_ctr,
                                     voxel_axon_dist=self.sd.axon_soma_dist,
                                     neuron=neuron_clones[neuron_id],
                                     neuron_id=neuron_id)

    def test_voxel_counts(self):

        if self.sd.voxel_overflow_counter > 0:
            print(f"Warning voxel overflow counter: {self.sd.voxel_overflow_counter}")

        dend_neuron_id, dend_count = np.unique(self.sd.dend_voxels.flatten(), return_counts=True)
        axon_neuron_id, axon_count = np.unique(self.sd.axon_voxels.flatten(), return_counts=True)

        min_dend_count = np.min(dend_count[1:])
        min_axon_count = np.min(axon_count[1:])
        max_dend_count = np.max(dend_count[1:])
        max_axon_count = np.max(axon_count[1:])

        dend_bins = np.arange(min_dend_count-1.5, max_dend_count+1.5)
        axon_bins = np.arange(min_axon_count-1.5, max_axon_count+1.5)

        if True:
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots(2, 1)
            ax[0].hist(dend_count[1:], bins=dend_bins, align="mid", color="black")  # 0 is empty elements
            ax[1].hist(axon_count[1:], bins=axon_bins, align="mid", color="blue")

            ax[0].plot(np.full((2,), dend_count[1]), ax[0].get_ylim(), color="red")
            ax[1].plot(np.full((2,), axon_count[1]), ax[1].get_ylim(), color="red")

            ax[0].plot(np.full((2,), np.mean(dend_count[1:])), ax[0].get_ylim(), color="black")
            ax[1].plot(np.full((2,), np.mean(axon_count[1:])), ax[1].get_ylim(), color="blue")

            ax[1].set_xlabel("Number of voxels")
            ax[0].set_ylabel("Count")
            ax[1].set_ylabel("Count")
            ax[0].set_title(f"Unrotated dendrite voxel count: {dend_count[1]} (spread {np.std(dend_count[1:]):.1f} relspread {np.std(dend_count[1:])/np.mean(dend_count[1:]):.5f} )")
            ax[1].set_title(f"Unrotated axon voxel count: {axon_count[1]} (spread {np.std(axon_count[1:]):.1f}, relspread {np.std(axon_count[1:])/np.mean(axon_count[1:]):.5f}))")

            fig.suptitle(f"Step multiplier: {self.sd.step_multiplier} (overflow: {self.sd.voxel_overflow_counter})")

            plt.tight_layout()

            if not os.path.isdir("figures"):
                os.mkdir("figures")

            fig_name = f"figures/detect_voxel_count-step-multiplier-{self.sd.step_multiplier}-num-{self.n_iter}.png"
            plt.savefig(fig_name)

            # plt.ion()
            #   plt.show()

            if False:
                print("Plotting hyper voxel (this is slow)")
                self.sd.plot_hyper_voxel()
                import pdb
                pdb.set_trace()

        self.assertTrue(np.std(dend_count[1:])/np.mean(dend_count[1:]) < 0.05)
        self.assertTrue(np.std(axon_count[1:])/np.mean(axon_count[1:]) < 0.05)

        # import pdb
        # pdb.set_trace()


if __name__ == '__main__':
    unittest.main()



import numpy as np


def deep_compare(a, b, float_tol=1e-9):
    """Recursively compares two structures with support for NumPy arrays and floating point tolerance."""

    if isinstance(a, dict) and isinstance(b, dict):
        if a.keys() != b.keys():
            return False
        return all(deep_compare(a[k], b[k], float_tol) for k in a)

    elif isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)):
        if len(a) != len(b) or type(a) != type(b):
            return False
        return all(deep_compare(x, y, float_tol) for x, y in zip(a, b))

    elif isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return a.shape == b.shape and np.allclose(a, b, atol=float_tol)

    elif isinstance(a, float) and isinstance(b, float):
        return abs(a - b) <= float_tol

    else:
        return a == b



import unittest
import os
import json

from snudda.init.init import SnuddaInit


class TestInit(unittest.TestCase):

    def setUp(self):

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

    def test_init(self):
        
        with self.subTest(stage="init_cube"):

            network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_init_cube")
    
            config_name = os.path.join(network_path, "network-config.json")
            snudda_data = os.path.join(os.path.dirname(__file__), "..", "snudda", "data")
            cnc = SnuddaInit(struct_def={}, config_file=config_name, snudda_data=snudda_data)
            cnc.define_striatum(num_dSPN=47500, num_iSPN=47500, num_FS=1300, num_LTS=0, num_ChIN=0,
                                volume_type="cube")
            cnc.write_json()

            with open(config_name, "r") as f:
                cfg_data = json.load(f)

            self.assertTrue("regions" in cfg_data)
            self.assertTrue("Striatum" in cfg_data["regions"])
            self.assertTrue("neurons" in cfg_data["regions"]["Striatum"])

            self.assertTrue("dSPN" in cfg_data["regions"]["Striatum"]["neurons"])
            self.assertEqual(cfg_data["regions"]["Striatum"]["neurons"]["dSPN"]["num_neurons"], 47500)

            self.assertTrue("iSPN" in cfg_data["regions"]["Striatum"]["neurons"])
            self.assertEqual(cfg_data["regions"]["Striatum"]["neurons"]["iSPN"]["num_neurons"], 47500)

            self.assertTrue("FS" in cfg_data["regions"]["Striatum"]["neurons"])
            self.assertEqual(cfg_data["regions"]["Striatum"]["neurons"]["FS"]["num_neurons"], 1300)

            self.assertTrue("LTS" not in cfg_data["regions"]["Striatum"]["neurons"])
            self.assertTrue("ChIN" not in cfg_data["regions"]["Striatum"]["neurons"])

            self.assertTrue(os.path.isfile(cfg_data["regions"]["Striatum"]["volume"]["mesh_file"]))
            self.assertEqual(cfg_data["regions"]["Striatum"]["volume"]["d_min"], 1.5e-5)

            self.assertTrue("connectivity" in cfg_data["regions"]["Striatum"])
            self.assertTrue("FS,FS" in cfg_data["regions"]["Striatum"]["connectivity"])
            self.assertTrue("FS,dSPN" in cfg_data["regions"]["Striatum"]["connectivity"])
            self.assertTrue("FS,iSPN" in cfg_data["regions"]["Striatum"]["connectivity"])
            self.assertTrue("dSPN,dSPN" in cfg_data["regions"]["Striatum"]["connectivity"])
            self.assertTrue("iSPN,iSPN" in cfg_data["regions"]["Striatum"]["connectivity"])
            self.assertTrue("iSPN,dSPN" in cfg_data["regions"]["Striatum"]["connectivity"])
            self.assertTrue("dSPN,iSPN" in cfg_data["regions"]["Striatum"]["connectivity"])

            self.assertTrue("GABA" in cfg_data["regions"]["Striatum"]["connectivity"]["FS,FS"])
            self.assertTrue("gap_junction" in cfg_data["regions"]["Striatum"]["connectivity"]["FS,FS"])

            self.assertTrue("conductance" in cfg_data["regions"]["Striatum"]["connectivity"]["FS,FS"]["GABA"])
            self.assertTrue("channel_parameters" in cfg_data["regions"]["Striatum"]["connectivity"]["FS,FS"]["GABA"])
            self.assertTrue("pruning" in cfg_data["regions"]["Striatum"]["connectivity"]["FS,FS"]["GABA"])

        with self.subTest(stage="init_slice"):
            network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_init_slice")

            config_name = os.path.join(network_path)
            cnc = SnuddaInit(struct_def={}, network_path=network_path)
            cnc.define_striatum(num_dSPN=47500, num_iSPN=47500, num_FS=1300, num_LTS=0, num_ChIN=0,
                                volume_type="slice")
            cnc.write_json()
            
        with self.subTest(stage="init_full"):
            network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_init_full")

            config_name = os.path.join(network_path, "network-config.json")
            cnc = SnuddaInit(struct_def={}, config_file=config_name, random_seed=1237)
            cnc.define_striatum(num_neurons=1670000)
            cnc.write_json()

            with open(config_name, "r") as f:
                cfg_data = json.load(f)

            self.assertEqual(cfg_data["random_seed"]["master_seed"], 1237)

        with self.subTest(stage="population-unit-random"):
            network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_init_pop_unit_random")
            cnc = SnuddaInit(struct_def={}, network_path=network_path)
            cnc.define_striatum(num_dSPN=1000, num_iSPN=1000, num_FS=20, num_LTS=0, num_ChIN=0,
                                volume_type="slice")
            cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["dSPN", "iSPN"],
                                           fraction_of_neurons=0.5)
            cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["dSPN", "iSPN"],
                                           fraction_of_neurons=0.2)
            cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["dSPN"],
                                           fraction_of_neurons=0.3)
            cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["iSPN"],
                                           fraction_of_neurons=0.15)
            cnc.add_population_unit_random(structure_name="Striatum", neuron_types=["iSPN"],
                                           fraction_of_neurons=0.15, unit_id=10)
            cnc.write_json()

            config_name = os.path.join(network_path, "network-config.json")
            with open(config_name, "r") as f:
                cfg_data = json.load(f)

            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["unit_id"] == [1, 2, 3, 4, 10])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["method"] == "random")
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["fraction_of_neurons"] == [0.5, 0.2, 0.3, 0.15, 0.15])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["unit_id"] == [1, 2, 3, 4, 10])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][0] == ["dSPN", "iSPN"])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][1] == ["dSPN", "iSPN"])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][2] == ["dSPN"])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][3] == ["iSPN"])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][4] == ["iSPN"])




        with self.subTest(stage="population-unit-sphere"):
            network_path = os.path.join(os.path.dirname(__file__), "networks", "network_testing_init_pop_unit_sphere")
            cnc = SnuddaInit(struct_def={}, network_path=network_path)
            cnc.define_striatum(num_dSPN=10000, num_iSPN=10000, num_FS=200, num_LTS=0, num_ChIN=0,
                                volume_type="slice")

            cnc.add_population_unit_density(structure_name="Striatum",
                                            neuron_types=["dSPN", "iSPN"],
                                            unit_centre=[0, 0, 0],
                                            probability_function="(d < 100e-6)*1")

            cnc.add_population_unit_density(structure_name="Striatum",
                                            neuron_types=["dSPN", "iSPN"],
                                            unit_centre=[300e-6, 0, 0],
                                            probability_function="exp(-d/200e-6)")

            cnc.add_population_unit_density(structure_name="Striatum",
                                            neuron_types=["dSPN"],
                                            unit_centre=[0, 300e-6, 0],
                                            probability_function="exp(-d/150e-6)",
                                            num_neurons=100)

            cnc.write_json()

            config_name = os.path.join(network_path, "network-config.json")
            with open(config_name, "r") as f:
                cfg_data = json.load(f)

            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["unit_id"] == [1, 2, 3])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["method"] == "radial_density")
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["centres"][0] == [0, 0, 0])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["centres"][1] == [0.000300, 0, 0])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["centres"][2] == [0, 0.000300, 0])

            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][0] == ["dSPN", "iSPN"])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][1] == ["dSPN", "iSPN"])
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["neuron_types"][2] == ["dSPN"])

            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["probability_functions"][0] == "(d < 100e-6)*1")
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["probability_functions"][1] == "exp(-d/200e-6)")
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["probability_functions"][2] == "exp(-d/150e-6)")

            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["num_neurons"][0] is None)
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["num_neurons"][1] is None)
            self.assertTrue(cfg_data["regions"]["Striatum"]["population_units"]["num_neurons"][2] == 100)




""" Based on BluePyOpt exampel code by Werner van Geit, 
modified by Johannes Hjorth """

import json
import os
from collections import OrderedDict
import numpy as np

import bluepyopt.ephys as ephys

from snudda.neurons.neuron_prototype import NeuronPrototype

# Fix for axon diameters, need to refactor in the future, or integrate directly in bluepyopt
from snudda.neurons.NrnFileMorphology_axon_fix import NrnFileMorphology_axon_fix


class NeuronModel(ephys.models.CellModel):
    """ Extended NeuronModel for simulation.

        Note that the paths are here assumed to be absolute, ie SNUDDA_DATA is not replaced, that should be done
        before calling this class.

    """

    def __init__(self,
                 cell_name="Unknown",
                 morph_path=None,
                 mech_file=None,
                 param_file=None,
                 modulation_file=None,
                 reaction_diffusion_file=None,
                 parameter_id=None,
                 morphology_id=None,
                 modulation_id=None,
                 parameter_key=None,
                 morphology_key=None,
                 modulation_key=None,
                 use_rxd_neuromodulation=True,
                 replace_axon_length=60e-6,
                 replace_axon_nseg_frequency=40e-6,
                 replace_axon_diameter=None,  # If this is not None, it uses special code to get diameter also
                 replace_axon_myelin_length=None,
                 replace_axon_myelin_diameter=None,
                 position=None,
                 rotation=None,
                 volume_id=None,
                 verbose=False):

        """
        Constructor

        Args:
            cell_name: Name of neuron, e.g. "dSPN"
            morph_path: Path to morphology directory, there may be multiple morphologies available
            mech_file: Path to mechanism file
            param_file: Path to parameter file
            modulation_file: Path to neuromodulation parameter file
            reaction_diffusion_file: Path to the RxD reaction diffusion file
            parameter_id: ID of parameter set
            morphology_id: ID of morphology set -- DEPRECATED
            modulation_id: ID of neuromodulation parameter set
            parameter_key (str): parameter key for lookup in parameter.json
            morphology_key (str): morphology key, together with parameter_key lookup in meta.json
            modulation_key (str): modulation key, lookup in modulation.json
            volume_id (str): name of volume (region) that neuron is located in

        """

        # OBS, modulation_id is not used. morphology_id, parameter_id and modulation_id will be deprecated
        # at some point. Key usage is preferred (less risk of mixing models)

        self.name = cell_name
        self.type = cell_name.split('_')[0]
        self.parameters = []

        self.script_dir = os.path.dirname(__file__)
        self.config_dir = os.path.join(self.script_dir, 'config')
        self.use_rxd_neuromodulation = use_rxd_neuromodulation

        self.is_relocated = False  # NEURON coordinates match Snudda simulation coordinates

        self.position = position
        self.rotation = rotation
        self.volume_id = volume_id

        self.verbose = verbose

        if os.path.isfile(morph_path):
            # If morph_path is a swc file, use it directly
            morph_file = morph_path
        else:
            # Figure out what file in the morphology path we should use

            morph_file = None
            if parameter_key is not None and morphology_key is not None and param_file is not None:
                meta_file = os.path.join(os.path.dirname(param_file), "meta.json")
                if os.path.isfile(meta_file):
                    with open(meta_file, "r") as f:
                        meta_data = json.load(f, object_pairs_hook=OrderedDict)
                        morph_file = os.path.join(morph_path, meta_data[parameter_key][morphology_key]["morphology"])

            if not morph_file:
                # We now allow multiple variations of morphologies for a given neuron name, so here NeuronPrototype
                # is used to acquire the actual morphology file we will use for this particular neuron
                # based on parameter_id and morphology_id.
                neuron_prototype = NeuronPrototype(neuron_name=cell_name,
                                                   neuron_path=None,
                                                   morphology_path=morph_path,
                                                   parameter_path=param_file,
                                                   mechanism_path=mech_file,
                                                   modulation_path=modulation_file,
                                                   reaction_diffusion_path=reaction_diffusion_file)

                morph_file, _ = neuron_prototype.get_morphology(parameter_id=parameter_id,
                                                                morphology_id=morphology_id,
                                                                parameter_key=parameter_key,
                                                                morphology_key=morphology_key)

            assert morph_file, (f"Neuron {cell_name} with morph_path = {morph_path} ({morphology_id}, "
                                f"parameter_path = {param_file} ({parameter_id}) "
                                f"has morph_file = {morph_file} (Should not be None)")

        self.morph_file = morph_file

        morph = self.define_morphology(replace_axon=True, morph_file=morph_file,
                                       replace_axon_length=replace_axon_length,
                                       replace_axon_nseg_frequency=replace_axon_nseg_frequency,
                                       replace_axon_diameter=replace_axon_diameter,
                                       replace_axon_myelin_length=replace_axon_myelin_length,
                                       replace_axon_myelin_diameter=replace_axon_myelin_diameter)

        mechs = self.define_mechanisms(mechanism_config=mech_file)
        params = self.define_parameters(param_file, parameter_id, parameter_key)

        if modulation_file:
            if modulation_key:
                mod_params = self.define_parameters(parameter_config=modulation_file,
                                                    parameter_key=modulation_key)
                params = params + mod_params
            else:
                print(f"Warning! No modulation key specified, ignoring {modulation_file}")

        super(NeuronModel, self).__init__(name=cell_name, morph=morph,
                                          mechs=mechs, params=params)

        if reaction_diffusion_file and self.use_rxd_neuromodulation:
            # Only load the module if actually used, this avoids weird RxD stuff happening in NEURON when not needed
            from snudda.neurons.neuron_modulation import NeuronModulation

            self.modulation = NeuronModulation(neuron=self, verbose=self.verbose)
            self.modulation.config_file = reaction_diffusion_file
        else:
            self.modulation = None

        self.syn_list = []
        self.section_lookup = None

    ##############################################################################

    # Helper function

    def define_mechanisms(self, mechanism_config=None):
        """Define mechanisms based on mechanism_config """

        assert (mechanism_config is not None)
        # print("Using mechanmism config: " + mechanism_config)

        with open(mechanism_config, "r") as f:
            mech_definitions = json.load(f, object_pairs_hook=OrderedDict)

        if "modpath" in mech_definitions:
            mod_path = os.path.join(self.script_dir, mech_definitions["modpath"])
            print(f"mod_path set to {mod_path} (not yet implemented)")
        else:
            mod_path = None

        mechanisms = []
        for section_list in mech_definitions:
            channels = mech_definitions[section_list]

            # This allows us to specify a modpath in the file
            if section_list == "modpath":
                continue

            seclist_loc = \
                ephys.locations.NrnSeclistLocation(section_list, seclist_name=section_list)
            for channel in channels:
                mechanisms.append(ephys.mechanisms.NrnMODMechanism(
                    name='%s.%s' % (channel, section_list),
                    mod_path=mod_path,
                    suffix=channel,
                    locations=[seclist_loc],
                    preloaded=True))

        return mechanisms

    ############################################################################

    # Helper function

    def define_parameters(self, parameter_config, parameter_id=None, parameter_key=None):
        """
        Define parameters based on parameter_config and parameter_id.
        If there are n parameter sets, and parameter_id is k, then the parameter set is n % k."""

        assert parameter_config is not None

        # print("Using parameter config: " + parameter_config)

        with open(parameter_config, "r") as f:
            param_configs = json.load(f, object_pairs_hook=OrderedDict)

        parameters = []

        if isinstance(param_configs, (OrderedDict, dict)):
            # Multiple parameters, pick one

            if parameter_key is not None:
                assert parameter_key in param_configs, f"Missing parameter key {parameter_key} in {parameter_config}"
                p_config = param_configs[parameter_key]

                if parameter_id is not None:
                    par_key_list = list(param_configs.keys())

                    par_key = par_key_list[parameter_id % len(par_key_list)]
                    assert par_key == parameter_key, \
                        (f"parameter_key mismatch, got {parameter_key}, but parameter_id={parameter_id} "
                         f"implies param_key={par_key}")
            else:
                assert parameter_id is not None, "Multiple parameter sets require parameter_id set"

                par_key_list = list(param_configs.keys())
                par_key = par_key_list[parameter_id % len(par_key_list)]
                p_config = param_configs[par_key]

        elif type(param_configs) == list and type(param_configs[0]) == list:
            # Parameter ID no longer exists, we default to parameter_id = 0
            # This is old fallback code, for old version format of parameters.json, remove in the future.
            if self.verbose:
                print("Warning: Old format of parameter config, using parameter_id = 0.")
            parameter_id = 0
            num_params = len(param_configs)
            p_config = param_configs[parameter_id % num_params]
        else:
            p_config = param_configs

        # Save this to be accessible in the future
        self.parameters += p_config

        for param_config in p_config:
            if 'value' in param_config:
                frozen = True
                value = param_config['value']
                bounds = None
            elif 'bounds' in param_config:
                frozen = False
                bounds = param_config['bounds']
                value = None
            else:
                raise Exception(f"Parameter config has to have bounds or value: {param_config}")

            if param_config['type'] == 'global':
                parameters.append(
                    ephys.parameters.NrnGlobalParameter(
                        name=param_config['param_name'],
                        param_name=param_config['param_name'],
                        frozen=frozen,
                        bounds=bounds,
                        value=value))
            elif param_config['type'] in ['section', 'range']:

                if param_config['dist_type'] == 'uniform':
                    scaler = ephys.parameterscalers.NrnSegmentLinearScaler()
                elif param_config['dist_type'] in ['exp', 'distance']:
                    scaler = ephys.parameterscalers.NrnSegmentSomaDistanceScaler(distribution=param_config['dist'])
                else:
                    raise ValueError(f"Unknown dist_type = {param_config['dist_type']}, "
                                     f"expected 'uniform', 'exp' or 'distance'")
                # 2024-07-23: Updated format, so that "sectionlist" is allowed to be either a string (of one section type)
                #             or a list of strings with section types.

                section_list = param_config['sectionlist']
                if not isinstance(section_list, list):
                    section_list = [section_list]

                for seclist_item in section_list:

                    seclist_loc = ephys.locations.NrnSeclistLocation(seclist_item,
                                                                     seclist_name=seclist_item)

                    name = '%s.%s' % (param_config['param_name'],
                                      seclist_item)

                    if param_config['type'] == 'section':
                        parameters.append(
                            ephys.parameters.NrnSectionParameter(
                                name=name,
                                param_name=param_config['param_name'],
                                value_scaler=scaler,
                                value=value,
                                frozen=frozen,
                                bounds=bounds,
                                locations=[seclist_loc]))
                    elif param_config['type'] == 'range':
                        parameters.append(
                            ephys.parameters.NrnRangeParameter(
                                name=name,
                                param_name=param_config['param_name'],
                                value_scaler=scaler,
                                value=value,
                                frozen=frozen,
                                bounds=bounds,
                                locations=[seclist_loc]))
            else:
                raise Exception(f"Param config type has to be global, section or range: {param_config}")

        return parameters

    ##############################################################################

    # Helper function

    def define_morphology(self, replace_axon=True, morph_file=None,
                          replace_axon_length=60e-6,
                          replace_axon_nseg_frequency=40e-6,
                          replace_axon_diameter=None,
                          replace_axon_myelin_length=None,
                          replace_axon_myelin_diameter=None):
        """
        Define morphology. Handles SWC and ASC.

        Args:
            replace_axon (bool): Is axon replaced with a stump for simulation, default True
            morph_file (str): Path to morphology
        """

        assert (morph_file is not None)

        if replace_axon_diameter is None and replace_axon_myelin_length is None:
            nrn_morph = ephys.morphologies.NrnFileMorphology(morph_file, do_replace_axon=replace_axon,
                                                             axon_stub_length=replace_axon_length*1e6,
                                                             axon_nseg_frequency=replace_axon_nseg_frequency*1e6)
        else:
            # Special treatment, if we have both length and diameter requirements
            if self.verbose:
                print(f"Axon diameter given, assuming both axon diameter and length are lists or arrays")

            nrn_morph = NrnFileMorphology_axon_fix(morphology_path=morph_file,
                                                   axon_length=replace_axon_length,
                                                   axon_diameter=replace_axon_diameter,
                                                   axon_nseg_frequency=replace_axon_nseg_frequency,
                                                   replace_axon_myelin_length=replace_axon_myelin_length,
                                                   replace_axon_myelin_diameter=replace_axon_myelin_diameter)

        return nrn_morph

    ##############################################################################

    # Neuron_morphology defines sectionID, these must match what this returns
    # so that they point to the same compartment.
    #
    # Soma is -1
    # axons are negative values (-2, -3, ..) in Neuron_morphology
    # dendrites are 0,1,2,3,4,5... ie one higher than what Neuron internally
    # uses to index the dendrites (due to us wanting to include soma)

    def build_section_lookup(self):

        self.section_lookup = dict([])

        # Soma is -1
        self.section_lookup[-1] = self.icell.soma[0]

        for ic, c in enumerate(self.icell.dend):
            self.section_lookup[ic] = c

        # Negative numbers for axon
        for ic, c in enumerate(self.icell.axon):
            self.section_lookup[-ic - 2] = c

    def map_id_to_compartment(self, section_id):

        """
        Map section_id to compartment.

        Neuron_morphology defines sectionID, these must match what this returns
        so that they point to the same compartment.

        Soma is -1
        axons are negative values (currently all set to -2) in Neuron_morphology
        dendrites are 0, 1,2,3,4,5...

        """

        if self.section_lookup is None:
            self.build_section_lookup()

        if isinstance(section_id, int):
            sec = self.section_lookup[section_id]
        else:
            sec = [self.section_lookup[x] for x in section_id]

        return sec

    # OVERRIDE the create_empty_template from CellModel
    '''create an hoc template named template_name for an empty cell'''

    @staticmethod
    def create_empty_template(
            template_name,
            seclist_names=None,
            secarray_names=None):

        objref_str = 'objref this, CellRef, synlist'
        new_seclist_str = ''

        if seclist_names:
            for seclist_name in seclist_names:
                objref_str += ', %s' % seclist_name
                new_seclist_str += \
                    '       %s = new SectionList()\n' % seclist_name

        create_str = ''
        if secarray_names:
            create_str = 'create '
            create_str += ', '.join(
                '%s[1]' % secarray_name
                for secarray_name in secarray_names)
            create_str += '\n'

        template = '''\
    begintemplate %(template_name)s
      %(objref_str)s
      proc init() {\n%(newseclist_str)s
      forall delete_section()
      CellRef = this
      synlist = new List()
      }
      gid = 0
      proc destroy() {localobj nil
      CellRef = nil
      }
      %(create_str)s
    endtemplate %(template_name)s
         ''' % dict(template_name=template_name, objref_str=objref_str,
                    newseclist_str=new_seclist_str,
                    create_str=create_str)

        return template

    ############################################################################

    # OVERRIDE the create_empty_template from CellModel
    """Create an empty cell in Neuron"""

    @staticmethod
    def create_empty_cell(
            name,
            sim,
            seclist_names=None,
            secarray_names=None):

        # TODO minize hardcoded definition
        # E.g. sectionlist can be procedurally generated
        # hoc_template = ephys.models.CellModel.create_empty_template(name,
        #                              seclist_names,
        #                              secarray_names)

        hoc_template = NeuronModel.create_empty_template(name,
                                                         seclist_names,
                                                         secarray_names)

        sim.neuron.h(hoc_template)

        template_function = getattr(sim.neuron.h, name)

        return template_function()

        ############################################################################

    # OVERRIDE the create_empty_template from CellModel

    def instantiate(self, sim=None, extracellular_regions=None):
        """Instantiate model in simulator"""

        # TODO replace this with the real template name
        if not hasattr(sim.neuron.h, self.name):
            self.icell = self.create_empty_cell(
                self.name,
                sim=sim,
                seclist_names=self.seclist_names,
                secarray_names=self.secarray_names)
        else:
            self.icell = getattr(sim.neuron.h, self.name)()

        self.icell.gid = self.gid

        self.morphology.instantiate(sim=sim, icell=self.icell)
        self.relocate_NEURON(sim=sim)

        for mechanism in self.mechanisms:
            mechanism.instantiate(sim=sim, icell=self.icell)
        for param in self.params.values():
            param.instantiate(sim=sim, icell=self.icell)

        if self.modulation:
            # TODO: THIS IS NOT THE Simulation object but the NEURON sim object... cant use it!!
            self.modulation.load_json(extracellular_regions=extracellular_regions, neuron_region=self.volume_id)

    ############################################################################

    def get_replacement_axon(self, axon_length=60e-6, axon_diameter=None, axon_nseg=None):

        raise DeprecationWarning("This function is now orphaned. Might need to re-update it if we switch to using hoc")

        """
            If axon_length is given as a scalar, then the code is similar to the BluePyOpt hoc, with the modification
            that the total axon_length is specified by the user.

            If axon_length is a vector, then axon_diameter and axon_nseg must be given and be vectors of same length
            The returned hoc will then create the user specified axon stump.

        """

        if isinstance(axon_length, (int, float, np.integer, np.floating)):
            assert axon_diameter is None and axon_nseg is None
            # We have only the length specified, use two sections

            assert axon_length < 10000e-6, "Please specify replacement axon_length in SI units (meters)."

            axon_length_specified_hoc = \
                f'''
            proc replace_axon(){{ local nSec, D1, D2
              // preserve the number of original axonal sections
              nSec = sec_count(axonal)

              // Try to grab info from original axon
              if(nSec == 0) {{ //No axon section present
                D1 = D2 = 1
              }} else if(nSec == 1) {{
                axon[0] D1 = D2 = diam
              }} else {{
                axon[0] D1 = D2 = diam
                soma distance() //to calculate distance from soma
                forsec axonal{{
                  //if section is longer than 60um then store diam and exit from loop
                  if(distance(0.5) > {axon_length * 1e6}){{
                    D2 = diam
                    break
                  }}
                }}
              }}

              // get rid of the old axon
              forsec axonal{{
                delete_section()
              }}

              create axon[2]

              axon[0] {{
                L = {axon_length * 1e6 / 2}
                diam = D1
                nseg = 1 + 2*int(L/40)
                all.append()
                axonal.append()
              }}
              axon[1] {{
                L = {axon_length * 1e6 / 2}
                diam = D2
                nseg = 1 + 2*int(L/40)
                all.append()
                axonal.append()
              }}
              nSecAxonal = 2
              soma[0] connect axon[0](0), 1
              axon[0] connect axon[1](0), 1
            }}
                    '''

            return axon_length_specified_hoc

        assert axon_length is not None and axon_diameter is not None and axon_nseg is not None

        # In all remaining cases the user has to specify vectors for axon_length, axon_diameter, axon_nseg
        assert len(axon_length) == len(axon_diameter) == len(axon_nseg), \
            f"Unequal lengths: {axon_length = }, {axon_diameter = }, {axon_nseg = }"

        user_defined_axon_hoc = \
            f"""
        proc replace_axon() {{
            
          // get rid of the old axon
          forsec axonal{{
            delete_section()
            
          create axon[{len(axon_length)}]
            
          }}
            """

        for idx, (al, ad, an) in enumerate(zip(axon_length, axon_diameter, axon_nseg)):

            assert 0 < al < 500e-6, "Please make sure you specify axon length in SI units (meters)"
            assert 0 < ad < 10e-6, "Please make sure you specify axon diameter in SI units (meters)"
            assert an % 2 == 1, f"{axon_nseg = } must contain odd integer"

            user_defined_axon_hoc += \
            f"""
          axon[{idx}] {{
            L = {al*1e6}
            diam = {ad*1e6}
            nseg = {an}
            all.append()
            axonal.append()
          }}
          """

        user_defined_axon_hoc += f"""
              nSecAxonal = {len(axon_length)}
        """

        for idx, al in enumerate(axon_length):

            if idx == 0:
                user_defined_axon_hoc += f"""
              soma[0] connect axon[0](0), 1
            """
            else:
                user_defined_axon_hoc += f"""
              axon[{idx-1}] connect axon[{idx}](0), 1
                """

        user_defined_axon_hoc += "}"

        return user_defined_axon_hoc

    def relocate_NEURON(self, sim):

        # TODO: Work in progress

        """ This function can only be called once per neuron, it updates the NEURON coordinates
            to match the Snudda coordinates. """

        if self.rotation is None:
            self.rotation = np.eye(3)

        if self.position is None:
            # Only relocate if a position is given
            return

        assert not self.is_relocated, f"You can only relocate a neuron once"
        self.is_relocated = True

        # Save the old NEURON soma center, subtract it from all coordinates
        # then apply the rotation matrix, and save the NEURON compartment coordinates

        num_points = int(self.icell.soma[0].n3d())
        x_center = sum(self.icell.soma[0].x3d(i) for i in range(num_points)) / num_points
        y_center = sum(self.icell.soma[0].y3d(i) for i in range(num_points)) / num_points
        z_center = sum(self.icell.soma[0].z3d(i) for i in range(num_points)) / num_points

        for sec in self.icell.soma[0].wholetree():
            n_points = int(sec.n3d())
            sim.neuron.h.pt3dstyle(0, sec=sec)  # We don't want the points to translate when parent point is moved

            for i in range(n_points):
                center_pos = np.array([sec.x3d(i) - x_center, sec.y3d(i) - y_center, sec.z3d(i) - z_center])
                rot_pos = np.matmul(self.rotation, center_pos)   # Ska det vara T på pos eller resultatet=
                new_pos = rot_pos + self.position*1e6  # Translated
                sim.neuron.h.pt3dchange(i, new_pos[0], new_pos[1], new_pos[2], sec.diam3d(i), sec=sec)

        #  print(f"Relocated neuron to {self.position}, with rotation {self.rotation}")
        


import os

import numpy as np


def create_slice_mesh(file_name,
                      centre_point=np.array([0, 0, 0]),
                      x_len=500e-6,
                      y_len=500e-6,
                      z_len=150e-6,
                      description=None):
    """
    Create a slice mesh in wavefront format.

    Args:
        file_name (str): Path to output file
        centre_point (float, float, float) : Centre point
        x_len (float): Width along x-dimension
        y_len (float): Width along y-dimension
        z_len (float) : Width along z-dimension
        description (str) : Description appended to file header
    """

    mesh_dir = os.path.dirname(file_name)
    if mesh_dir and not os.path.exists(mesh_dir):
        os.makedirs(mesh_dir)

    # 2019-11-26 : Anya said that her sagital striatal slices were 2.36 x 2.36 mm.
    #              so that can be an upper limit
    if type(centre_point) is not np.ndarray:
        centre_point = np.array(centre_point)

    print("Creating slice mesh")
    print(f"File: {file_name}")
    print(f"Centre: {centre_point}")
    print(f"Sides: {x_len} x {y_len} x {z_len}")
    print(f"Description: {description}")

    vertex = np.array([[0.0, 0.0, 0.0],
                       [0.0, 0.0, 1.0],
                       [0.0, 1.0, 0.0],
                       [0.0, 1.0, 1.0],
                       [1.0, 0.0, 0.0],
                       [1.0, 0.0, 1.0],
                       [1.0, 1.0, 0.0],
                       [1.0, 1.0, 1.0]])

    # Centre cube
    vertex -= np.array([0.5, 0.5, 0.5])

    # Scale the cube to right size
    vertex[:, 0] *= x_len
    vertex[:, 1] *= y_len
    vertex[:, 2] *= z_len

    # Position cube

    vertex += centre_point

    vertex *= 1e6  # The other obj files are in micrometers, so convert :(

    normal_str = "vn  0.0  0.0  1.0\n" \
                 + "vn  0.0  0.0 -1.0\n" \
                 + "vn  0.0  1.0  0.0\n" \
                 + "vn  0.0 -1.0  0.0\n" \
                 + "vn  1.0  0.0  0.0\n" \
                 + "vn -1.0  0.0  0.0\n"

    face_str = "f  1//2  7//2  5//2\n" \
               + "f  1//2  3//2  7//2\n" \
               + "f  1//6  4//6  3//6\n" \
               + "f  1//6  2//6  4//6\n" \
               + "f  3//3  8//3  7//3\n" \
               + "f  3//3  4//3  8//3\n" \
               + "f  5//5  7//5  8//5\n" \
               + "f  5//5  8//5  6//5\n" \
               + "f  1//4  5//4  6//4\n" \
               + "f  1//4  6//4  2//4\n" \
               + "f  2//1  6//1  8//1\n" \
               + "f  2//1  8//1  4//1\n"

    with open(file_name, 'wt') as f:
        f.write("# Generated by create_slice_mesh.py\n")
        f.write(f"# {description}\n\n")
        f.write("g cube\n\n")

        for row in vertex:
            f.write("v %f %f %f\n" % tuple(row))

        f.write(f"\n{normal_str}")
        f.write(f"\n{face_str}")


if __name__ == "__main__":
    create_slice_mesh(file_name="test-slice-alex.obj",
                      centre_point=[0.5, 0.5, 0.0],
                      x_len=1e-3,
                      y_len=1e-3,
                      z_len=25e-6,
                      description="This is a test slice for Alex")



# Plan:
#
# Randomise dtau, ftau, U -- check that it is within the ranges specified
# for pair pulse ratio, and recovery test ratio
#
# Fit the neuron synapse model to the modelled traces
#
#

import json
import os

import numpy as np
import scipy.stats as stats

import matplotlib.pyplot as plt

from snudda.utils.numpy_encoder import NumpyEncoder

################################################################################


class Planert2010(object):

    def __init__(self, par_type, pre_num=1000000, max_num=100):

        self.con_info = None
        self.t_stim = None  # Set in def_exp_data

        self.def_exp_data()
        pars = self.pick_random_param(par_type=par_type, num=pre_num)
        amps = None

        if True:
            # Just checking that original starting distributions are ok
            self.plot_model_params(par_type, pars)

        amps = self.check_data_match(par_type=par_type, pars=pars, amps=amps)

        for ctr in range(0, 3):
            if False:
                # This tries to prune the distribution of the variables back to
                # the range
                pars, amps = self.prune_parameters(par_type=par_type, pars=pars, amps=amps)
                self.check_data_match(par_type=par_type, pars=pars, amps=amps)

            pars, amps = self.prune_data(par_type=par_type, pars=pars, amps=amps)
            self.check_data_match(par_type=par_type, pars=pars, amps=amps)
            self.plot_model_params(par_type, pars)

        pars, amps = self.max_num_params(pars, amps, max_num=max_num)

        self.save_data(par_type, pars, amps)

        self.check_data_match(par_type=par_type, pars=pars, amps=amps)
        self.plot_model_params(par_type, pars)

    ############################################################################

    def prune_data(self, par_type, pars, amps=None):

        if amps.shape[0] == 0:
            print("No data left?")
            import pdb
            pdb.set_trace()

        ppr_mean, ppr_std = self.con_info[par_type]["ppr"]
        rtr_mean, rtr_std = self.con_info[par_type]["rtr"]

        if amps is None:
            amps = self.synapse_model_wrapper(self.t_stim, pars)

        ppr = np.divide(amps[:, 1], amps[:, 0])
        rtr = np.divide(amps[:, -1], amps[:, 0])

        n_bins = 20
        ppr_max = np.max(ppr)
        rtr_max = np.max(rtr)

        # Which bin does each ppr belong to
        ppr_idx = (n_bins * ppr / (ppr_max + 1e-6)).astype(int)
        rtr_idx = (n_bins * rtr / (rtr_max + 1e-6)).astype(int)

        ppr_count = np.zeros((n_bins,))
        rtr_count = np.zeros((n_bins,))

        for p, r in zip(ppr_idx, rtr_idx):
            ppr_count[p] += 1
            rtr_count[r] += 1

        ppr_bin_width = ppr_max / (n_bins - 1)
        rtr_bin_width = rtr_max / (n_bins - 1)

        ppr_centre = ppr_bin_width / 2 + np.arange(0, n_bins) * ppr_bin_width
        rtr_centre = rtr_bin_width / 2 + np.arange(0, n_bins) * rtr_bin_width

        ppr_density = stats.norm.pdf(ppr_centre, ppr_mean, ppr_std)
        rtr_density = stats.norm.pdf(rtr_centre, rtr_mean, rtr_std)

        # We pick random param sets, and see if that PPR and RTR are over
        # represented, if so we remove it.

        n_left = len(pars["u"])
        keep_flag = np.ones((n_left,), dtype=bool)

        idx_rand = np.random.permutation(n_left)

        for idx in idx_rand:
            ppr_bin = ppr_idx[idx]
            rtr_bin = rtr_idx[idx]
            ppr_p = ppr_count[ppr_bin] / n_left / ppr_bin_width
            rtr_p = rtr_count[rtr_bin] / n_left / rtr_bin_width

            # if(pprP > pprDensity[pprBin] or rtrP > rtrDensity[rtrBin] \
            #   or pprBin == 0 or rtrBin == 0):
            if (ppr_p / ppr_density[ppr_bin] + rtr_p / rtr_density[rtr_bin] > 2
                    or ppr_bin == 0 or rtr_bin == 0
                    or ppr_density[ppr_bin] < 1e-4 or rtr_density[rtr_bin] < 1e-4):
                # This point is over represented, lets remove it
                keep_flag[idx] = False
                n_left -= 1
                ppr_count[ppr_idx[idx]] -= 1
                rtr_count[rtr_idx[idx]] -= 1

        pars["u"] = pars["u"][keep_flag]
        pars["ftau"] = pars["ftau"][keep_flag]
        pars["dtau"] = pars["dtau"][keep_flag]
        pars["amp"] = pars["amp"][keep_flag]

        print(f"(Peaks) Reduced {len(keep_flag)} down to {sum(keep_flag)}")

        return pars, amps[keep_flag, :]

    ############################################################################

    @staticmethod
    def max_num_params(pars, amps, max_num=100):

        n_par_sets = len(pars["u"])
        if n_par_sets < max_num:
            return pars, amps

        print(f"Reducing from {n_par_sets} down to {max_num} parameter sets")

        pars["u"] = pars["u"][:max_num]

        pars["ftau"] = pars["ftau"][:max_num]
        pars["dtau"] = pars["dtau"][:max_num]
        pars["amp"] = pars["amp"][:max_num]

        return pars, amps[:max_num, :]

        ############################################################################

    # This prunes the parameters to get back their distribution to match

    def prune_parameters(self, par_type, pars, amps):

        (d_tau_mean, d_tau_std) = self.con_info[par_type]["dtau"]
        (f_tau_mean, f_tau_std) = self.con_info[par_type]["ftau"]
        (u_mean, u_std) = self.con_info[par_type]["U"]

        n_bins = 20

        u_max = np.max(pars["u"])
        d_tau_max = np.max(pars["dtau"])
        f_tau_max = np.max(pars["ftau"])

        u_idx = (n_bins * pars["u"] / (u_max + 1e-6)).astype(int)
        d_tau_idx = (n_bins * pars["dtau"] / (d_tau_max + 1e-6)).astype(int)
        f_tau_idx = (n_bins * pars["ftau"] / (f_tau_max + 1e-6)).astype(int)

        u_count = np.zeros((n_bins,))
        d_tau_count = np.zeros((n_bins,))
        f_tau_count = np.zeros((n_bins,))

        for u, d, f in zip(u_idx, d_tau_idx, f_tau_idx):
            u_count[u] += 1
            d_tau_count[d] += 1
            f_tau_count[f] += 1

        u_bin_width = u_max / (n_bins - 1)
        d_tau_bin_width = d_tau_max / (n_bins - 1)
        f_tau_bin_width = f_tau_max / (n_bins - 1)

        u_centre = u_bin_width / 2 + np.arange(0, n_bins) * u_bin_width
        d_tau_centre = d_tau_bin_width / 2 + np.arange(0, n_bins) * d_tau_bin_width
        f_tau_centre = f_tau_bin_width / 2 + np.arange(0, n_bins) * f_tau_bin_width

        u_density = stats.norm.pdf(u_centre, u_mean, u_std)
        d_tau_density = stats.norm.pdf(d_tau_centre, d_tau_mean * 1e-3, d_tau_std * 1e-3)
        f_tau_density = stats.norm.pdf(f_tau_centre, f_tau_mean * 1e-3, f_tau_std * 1e-3)

        n_left = len(pars["u"])
        keep_flag = np.ones((n_left,), dtype=bool)

        idx_rand = np.random.permutation(n_left)

        for idx in idx_rand:
            u_bin = u_idx[idx]
            d_tau_bin = d_tau_idx[idx]
            f_tau_bin = f_tau_idx[idx]

            u_p = u_count[u_bin] / n_left / u_bin_width
            d_tau_p = d_tau_count[d_tau_bin] / n_left / d_tau_bin_width
            f_tau_p = f_tau_count[f_tau_bin] / n_left / f_tau_bin_width

            if (u_p / u_density[u_bin] + d_tau_p / d_tau_density[d_tau_bin] + f_tau_p / f_tau_density[f_tau_bin] > 3
                    or u_density[u_bin] < 1e-4 or d_tau_density[d_tau_bin] < 1e-4 or f_tau_density[f_tau_bin] < 1e-4):
                keep_flag[idx] = False
                n_left -= 1
                u_count[u_idx[idx]] -= 1
                d_tau_count[d_tau_idx[idx]] -= 1
                f_tau_count[f_tau_idx[idx]] -= 1

                # import pdb
                # pdb.set_trace()

        pars["u"] = pars["u"][keep_flag]
        pars["ftau"] = pars["ftau"][keep_flag]
        pars["dtau"] = pars["dtau"][keep_flag]
        pars["amp"] = pars["amp"][keep_flag]

        print(f"(Pars) Reduced {len(keep_flag)} down to {sum(keep_flag)}")

        return pars, amps[keep_flag, :]

    ############################################################################

    def check_data_match(self, par_type, pars, amps=None):

        ppr_mean, ppr_std = self.con_info[par_type]["ppr"]
        rtr_mean, rtr_std = self.con_info[par_type]["rtr"]

        if amps is None:
            amps = self.synapse_model_wrapper(self.t_stim, pars)

        ppr = np.divide(amps[:, 1], amps[:, 0])
        rtr = np.divide(amps[:, -1], amps[:, 0])

        # self.plotSpikes(amps,self.tSpike,pars,0)

        if True:
            self.plot_data_match(values=ppr,
                                 mean=ppr_mean,
                                 std=ppr_std,
                                 title=par_type,
                                 xlabel="PPR")

            self.plot_data_match(values=rtr,
                                 mean=rtr_mean,
                                 std=rtr_std,
                                 title=par_type,
                                 xlabel="RTR")

        return amps

    ############################################################################

    @staticmethod
    def plot_data_match(values, mean, std, title, xlabel, n_bins=20):

        print(f"{xlabel} Mean : {mean}, std : {std}")
        print(f"Data : {np.mean(values)}, std : {np.std(values)}")
        n_points = 100
        x = np.linspace(mean - 3 * std, mean + 3 * std, n_points)

        plt.figure()
        plt.plot(x, stats.norm.pdf(x, mean, std))
        plt.hist(values, density=True, bins=n_bins)
        plt.title(title)
        plt.xlabel(xlabel)
        plt.ion()
        plt.show()
        plt.tight_layout()

        fig_name = os.path.join("DATA", "Planert2010", "figs",
                                f"Planert2010-surrogate-data-pre-fit-{title}-{xlabel}.pdf")
        plt.savefig(fig_name)
        plt.close()

    ############################################################################

    @staticmethod
    def plot_spikes(amps, t_stim, pars=None, idx=0):

        plt.figure()
        for a, t in zip(amps[idx, :], t_stim):
            plt.plot([t, t], [0, a], color="black")
        plt.xlabel("Time")
        plt.ylabel("Amplitude")

        if pars is not None:
            title_str = "U = %.3f, ftau = %.3f, dtau = %.3f, amp = %.3f" \
                        % (pars["u"][idx], pars["ftau"][idx], pars["dtau"][idx], pars["amp"][idx])
            plt.title(title_str)

        plt.ion()
        plt.show()

    ############################################################################

    # parType = "DD", "DI", "ID", "II", "FD" or "FI"

    def pick_random_param(self, par_type, num):

        print(f"Picking {num} {par_type} parameter sets...")

        # Need to randomise amp, dtau, ftau and U

        amp_mean, amp_std = self.con_info[par_type]["amp"]
        d_tau_mean, d_tau_std = self.con_info[par_type]["dtau"]
        f_tau_mean, f_tau_std = self.con_info[par_type]["ftau"]
        u_mean, u_std = self.con_info[par_type]["U"]

        amp = (np.random.normal(size=num, scale=amp_std) + amp_mean) * 1e-3
        d_tau = (np.random.normal(size=num, scale=d_tau_std) + d_tau_mean) * 1e-3
        f_tau = (np.random.normal(size=num, scale=f_tau_std) + f_tau_mean) * 1e-3
        u = np.random.normal(size=num, scale=u_std) + u_mean

        # If any of the conditions are true, we are outside OK range,
        # so only OK if all conditions are false, and the result sums to False
        ok_idx = False == (np.array(amp < 0) + np.array(d_tau < 0)
                           + np.array(f_tau < 0) + np.array(u < 0) + np.array(u > 1))

        # Then we need to verify that the ppr and rtr are within range

        print(f"Npoints = {np.sum(ok_idx)}")

        pars = {"amp": amp[ok_idx],
                "dtau": d_tau[ok_idx],
                "ftau": f_tau[ok_idx],
                "u": u[ok_idx]}

        return pars

    ############################################################################

    def synapse_model_wrapper(self, t_stim, pars):

        print(f"Running {len(pars['u'])} simulations")

        amps = np.zeros((len(pars["u"]), len(t_stim)))

        for i, (u, d_tau, f_tau, Asc) in enumerate(zip(pars["u"],
                                                       pars["dtau"],
                                                       pars["ftau"],
                                                       pars["amp"])):
            if i > 0 and i % 100000 == 0:
                # Print progress...
                print(str(i) + "/" + str(amps.shape[0]))

            amps[i, :] = self.synapse_model(t_stim, U=u, d_tau=d_tau, f_tau=f_tau, asc=Asc)

        return amps

    ############################################################################

    @staticmethod
    def synapse_model(t_stim, U, d_tau, f_tau, asc):

        n_spikes = len(t_stim)
        u = np.zeros((n_spikes,))
        r = np.zeros((n_spikes,))
        isi = np.diff(t_stim)

        # Init
        u[0] = U
        r[0] = 1

        for idx, ii in enumerate(isi):
            u[idx + 1] = u[idx] * np.exp(-isi[idx] / f_tau) + U * (1 - u[idx] * np.exp(-isi[idx] / f_tau))

            r[idx + 1] = r[idx] * (1 - u[idx]) * np.exp(-isi[idx] / d_tau) + 1 - np.exp(-isi[idx] / d_tau)

        amp = np.multiply(u, r) * asc

        return amp

    ############################################################################

    def def_exp_data(self):

        self.con_info = dict()

        self.con_info["DD"] = {"P": (3, 43),
                               "amp": (0.24, 0.15),
                               "ppr": (0.91, 0.63),
                               "rtr": (1.23, 0.50),
                               "dtau": (192, 114),
                               "ftau": (1266, 1427),
                               "U": (0.39, 0.22),
                               "FDratio": 4.76}

        self.con_info["DI"] = {"P": (3, 66),
                               "amp": (0.33, 0.15),
                               "ppr": (0.84, 0.3),
                               "rtr": (1.16, 0.29),
                               "dtau": (96, 9),
                               "ftau": (313, 363),
                               "U": (0.46, 0.24),
                               "FDratio": 4.76}

        self.con_info["ID"] = {"P": (10, 80),
                               "amp": (0.27, 0.09),
                               "ppr": (1.1, 0.6),
                               "rtr": (1.51, 0.64),
                               "dtau": (365, 471),
                               "ftau": (570, 783),
                               "U": (0.36, 0.18),
                               "FDratio": 4.76}

        self.con_info["II"] = {"P": (7, 31),
                               "amp": (0.45, 0.44),
                               "ppr": (0.95, 0.48),
                               "rtr": (1.39, 0.69),
                               "dtau": (149, 90),
                               "ftau": (1462, 1800),
                               "U": (0.34, 0.19),
                               "FDratio": 4.76}

        self.con_info["FD"] = {"P": (8, 9),
                               "amp": (4.8, 4.9),
                               "ppr": (0.62, 0.12),
                               "rtr": (0.72, 0.08),
                               "dtau": (740, 350),
                               "ftau": (3.1, 2.4),
                               "U": (0.24, 0.07),
                               "FDratio": 0.16}

        self.con_info["FI"] = {"P": (6, 9),
                               "amp": (3.1, 4.1),
                               "ppr": (0.66, 0.14),
                               "rtr": (0.63, 0.19),
                               "dtau": (850, 500),
                               "ftau": (4.5, 2.7),
                               "U": (0.23, 0.07),
                               "FDratio": 0.16}

        self.t_stim = [0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.9]

    ############################################################################

    def save_data(self, par_type, pars, amp):
        file_name = os.path.join("DATA", "Planert2010", f"PlanertFitting-{par_type}-cache.json")

        data = dict()

        for p in pars:
            data[p] = pars[p]

        data["simAmp"] = amp
        data["tStim"] = self.t_stim

        with open(file_name, "w") as f:
            json.dump(data, f, indent=2, cls=NumpyEncoder)

    ############################################################################

    @staticmethod
    def load_data(par_type):

        file_name = os.path.join("Data", "Planert2010", f"PlanertFitting-{par_type}-cache.json")

        if not os.path.exists(file_name):
            return None, None

        with open(file_name, "r") as f:
            data = json.load(f)

        pars = dict([])
        for p in data:
            if p in ["u", "dtau", "ftau", "amp"]:
                pars[p] = data[p]
            elif p != "simAmp":
                print(f"Unknown parameter : {p}")

        amps = data["simAmp"]

        return pars, amps

    ############################################################################

    def plot_model_params(self, data_type, pars):

        (d_tau_mean, d_tau_std) = self.con_info[data_type]["dtau"]
        (f_tau_mean, f_tau_std) = self.con_info[data_type]["ftau"]
        (u_mean, u_std) = self.con_info[data_type]["U"]

        self._plot_model_params_helper(data_type, d_tau_mean * 1e-3, d_tau_std * 1e-3, pars["dtau"], "dtau")
        self._plot_model_params_helper(data_type, f_tau_mean * 1e-3, f_tau_std * 1e-3,
                                       pars["ftau"], "ftau")
        self._plot_model_params_helper(data_type, u_mean, u_std, pars["u"], "U")

        ############################################################################

    @staticmethod
    def _plot_model_params_helper(data_type, data_mean, data_std, data_points, x_label):

        x_min = np.min(data_points)
        x_max = np.max(data_points)

        x = np.linspace(x_min, x_max, 100)
        x_density = stats.norm.pdf(x, data_mean, data_std)  # / (x[1]-x[0])

        plt.figure()
        plt.hist(data_points, bins=20, density=True)
        plt.plot(x, x_density)
        plt.xlabel(x_label)
        plt.ylabel("Density")

        plt.tight_layout()

        fig_name = os.path.join("DATA", "Planert2010", "figs",
                                f"Surrogate-variables-distribution-{data_type}-{x_label}.pdf")
        plt.savefig(fig_name)
        plt.close()

    ############################################################################


if __name__ == "__main__":
    n_runs = 1000000
    max_num = 50

    pp = Planert2010(par_type="FI", pre_num=n_runs, max_num=max_num)
    pp = Planert2010(par_type="FD", pre_num=n_runs, max_num=max_num)

    pp = Planert2010(par_type="DD", pre_num=n_runs, max_num=max_num)
    pp = Planert2010(par_type="DI", pre_num=n_runs, max_num=max_num)
    pp = Planert2010(par_type="ID", pre_num=n_runs, max_num=max_num)
    pp = Planert2010(par_type="II", pre_num=n_runs, max_num=max_num)





import os
import numpy as np
import pandas as pd
import json
import quantities as pq


# See convert_sbtab_to_json.sh -- example


class ReadSBtab:

    def __init__(self, sbtab_path, out_filename,
                 compounds_filename=None,
                 reactions_filename=None,
                 parameters_filename=None,
                 constants_filename=None):

        self.sbtab_path = sbtab_path

        if compounds_filename is not None:
            self.compounds_filename = compounds_filename
        else:
            self.compounds_filename = "Compound.tsv"

        if reactions_filename is not None:
            self.reactions_filename = reactions_filename
        else:
            self.reactions_filename = "Reaction.tsv"

        if parameters_filename is not None:
            self.parameters_filename = parameters_filename
        else:
            self.parameters_filename = "Parameter.tsv"

        if constants_filename is not None:
            self.constants_filename = constants_filename
        else:
            self.constants_filename = "Constant.tsv"

        self.out_filename = out_filename

        self.compounds_data = None
        self.reactions_data = None
        self.parameters_data = None
        self.constants_data = None

        self.default_concentration_unit = "molar"
        self.unit_dict = dict()

        self.unit_dict["nmol_unit"] = pq.UnitQuantity('nanomole', pq.nano * pq.mole, symbol='nmol')
        self.unit_dict["nM_unit"] = pq.UnitQuantity('nanomolar', pq.nano * pq.molar, symbol='nM')
        self.unit_dict["liter"] = pq.UnitQuantity("liter", pq.litre, symbol="liter")

        self.parameters = dict()

        self.data = dict()

        self.load()

    def load(self):
        compound_path = os.path.join(self.sbtab_path, self.compounds_filename)
        reaction_path = os.path.join(self.sbtab_path, self.reactions_filename)
        parameter_path = os.path.join(self.sbtab_path, self.parameters_filename)
        constant_path = os.path.join(self.sbtab_path, self.constants_filename)

        for dest, source_file in [("compounds_data", compound_path),
                                  ("reactions_data", reaction_path),
                                  ("parameters_data", parameter_path),
                                  ("constants_data", constant_path)]:
            print(f"Loading {source_file}")
            data = pd.read_csv(source_file, sep="\t", skiprows=1)
            setattr(self, dest, data)

    def _get_optional_value(self, row, value_name, default_value=0):

        value = row.get(value_name, default_value)

        if value is None or value == np.nan:
            value = default_value

        return value

    def parse(self):

        self.data["species"] = dict()
        self.data["rates"] = dict()
        self.data["reactions"] = dict()

        original_sbtab_units = dict()

        nmol_unit = self.unit_dict["nmol_unit"]
        nM_unit = self.unit_dict["nM_unit"]

        # conc_unit = nM_unit
        conc_unit = pq.UnitQuantity("millimolar", pq.M * pq.milli, "mM")      # mM = mol / m3 = milli mol / litre

        # TODO: Should UNITS be nano Molar or Molar?

        for row_idx, row in self.compounds_data.iterrows():

            try:
                species_name = row["!Name"]
                species_unit = pq.CompoundUnit(row["!Unit"])  # str ==> enhet
                rescale_factor = float(species_unit.rescale(conc_unit).base)  # To get in nano molar
            except Exception as e:
                import traceback
                print(traceback.format_exc())
                print(e)
                import pdb
                pdb.set_trace()

            species_data = { "initial_concentration": row["!InitialValue"] * rescale_factor,
                             "diffusion_constant": self._get_optional_value(row, "DiffusionConstant", 0),
                             "charge": self._get_optional_value(row, "Charge", 0),
                             "regions": self._get_optional_value(row, "Regions", ["soma_internal", "dend_internal"]),
                             "concentration": row["!InitialValue"] * rescale_factor,
                             "boundary_condition": row["!IsConstant"] == 1,
                             "unit": conc_unit.symbol
                             }

            self.data["species"][species_name] = species_data
            original_sbtab_units[species_name] = species_unit.symbol 

        self.get_parameters()

        for row_idx, row in self.reactions_data.iterrows():
            reaction_name = row["!Name"]
            reactants, products = row["!ReactionFormula"].split("<=>")
            reactants = self._format_component_str(reactants)
            products = self._format_component_str(products)
            kinetic_law = row["!KineticLaw"].split("-")

            param_name_forward = kinetic_law[0].split("*")[0].strip()
            forward_rate = self.parameters[param_name_forward]["value"]
            forward_unit = self.parameters[param_name_forward]["unit"]

            if len(kinetic_law) == 2:
                param_name_backward = kinetic_law[1].split("*")[0].strip()
                backward_rate = self.parameters[param_name_backward]["value"]
                backward_unit = self.parameters[param_name_backward]["unit"]
            else:
                backward_rate = None
                backward_unit = None

            # TODO: Allow user to specify regions
            reaction_info = {"reactants": reactants,
                             "products": products,
                             "forward_rate": forward_rate,
                             "backward_rate": backward_rate,
                             "forward_rate_unit": forward_unit,
                             "backward_rate_unit": backward_unit,
                             "regions": ["soma_internal", "dend_internal"]}

            self.data["reactions"][reaction_name] = reaction_info

    def _format_component_str(self, component_str):

        new_str = ""
        for comp in component_str.strip().split(" "):
            new_str += f" {comp}"
            if comp.isdigit():
                new_str += " *"

        return new_str.strip()

    def get_parameters(self):

        for row_idx, row in self.parameters_data.iterrows():
            parameter_name = row["!Name"]
            parameter_value = row["!Value:linspace"]

            original_unit_str = row["!Unit"]

            nmol_unit = self.unit_dict["nmol_unit"]
            nM_unit = self.unit_dict["nM_unit"]

            original_unit = pq.CompoundUnit(original_unit_str)
            simplified_unit = original_unit.simplified
            target_unit_str = str(simplified_unit.units._dimensionality).replace("**", "^")
            scale = float(simplified_unit.base)

            print(f"{original_unit_str = }, {target_unit_str = }, {scale = }")

            try:
                self.parameters[parameter_name] = {"value": parameter_value * scale,
                                                   "unit": target_unit_str}
            except:
                import traceback
                print(traceback.format_exc())
                import pdb
                pdb.set_trace()

    def _get_rates(self, rate_str):
        split_str = rate_str.split("-")
        if len(split_str) == 1:
            forward_rate_name = split_str.split("*")[0]

    def write(self, out_file=None):

        if out_file is None:
            out_file = self.out_filename

        if out_file is None:
            raise ValueError(f"No outfile specified.")

        print(f"Writing JSON to {out_file}")

        with open(out_file, "wt") as f:
            json.dump(self.data, f, indent=4)

def cli():

    import argparse

    parser = argparse.ArgumentParser(description="Convert SBML to Snudda reaction diffusion with RxD",
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument("sbtab_path", help="SBtab path", type=str)

    parser.add_argument("--compounds_file", help="Input SBtab compound file to convert", default=None, type=str)
    parser.add_argument("--reactions_file", help="Input SBtab reaction file to convert", default=None, type=str)
    parser.add_argument("--parameters_file", help="Input SBtab parameters file to convert", default=None, type=str)
    parser.add_argument("--constants_file", help="Input SBtab constants file to convert", default=None, type=str)

    parser.add_argument("json_file", help="Snudda RxD JSON output file")

    args = parser.parse_args()

    rs = ReadSBtab(sbtab_path=args.sbtab_path,
                   compounds_filename=args.compounds_file,
                   reactions_filename=args.reactions_file,
                   parameters_filename=args.parameters_file,
                   constants_filename=args.constants_file,
                   out_filename=args.json_file)

    rs.parse()
    rs.write(args.json_file)


if __name__ == "__main__":
    cli()



import glob
import json
import os

from snudda.neurons import NeuronMorphologyExtended
from snudda.utils.snudda_path import snudda_parse_path


class NeuronPrototype:
    """ Helper class, returns a neuron prototype based on parameter_id, morph_id and modulation_id """

    def __init__(self,
                 neuron_path,
                 neuron_name,
                 morphology_path=None,
                 parameter_path=None,
                 mechanism_path=None,
                 modulation_path=None,
                 reaction_diffusion_path=None,
                 snudda_data=None,
                 meta_path=None,
                 virtual_neuron=False,
                 load_morphology=True,
                 verbose=False):

        self.verbose = verbose
        self.snudda_data = snudda_data

        if neuron_path:
            self.neuron_path = snudda_parse_path(neuron_path, self.snudda_data)
        elif parameter_path:
            self.neuron_path = os.path.dirname(snudda_parse_path(parameter_path, self.snudda_data))
        elif morphology_path:
            # morphology path usually points to "morphology" directory, but it can also point to a specific morphology
            # which should be in neuron_path then (old format), but it might be in morphology folder...

            if os.path.isdir(snudda_parse_path(morphology_path, self.snudda_data)):
                self.neuron_path = os.path.dirname(snudda_parse_path(morphology_path, self.snudda_data))
            elif os.path.isfile(snudda_parse_path(morphology_path, self.snudda_data)):
                morph_path = os.path.dirname(snudda_parse_path(morphology_path, self.snudda_data))
                if "morphology" in morph_path:
                    self.neuron_path = os.path.dirname(morph_path)
                else:
                    self.neuron_path = morph_path
        else:
            self.neuron_path = None

        if morphology_path:
            self.morphology_path = snudda_parse_path(morphology_path, self.snudda_data)
        elif self.neuron_path:
            self.morphology_path = snudda_parse_path(os.path.join(self.neuron_path, "morphology"), self.snudda_data)
        else:
            self.morphology_path = None

        if mechanism_path:
            self.mechanism_path = snudda_parse_path(mechanism_path, self.snudda_data)
        elif self.neuron_path:
            self.mechanism_path = snudda_parse_path(os.path.join(self.neuron_path, "mechanisms.json"), self.snudda_data)
        else:
            self.mechanism_path = None

        if parameter_path:
            self.parameter_path = snudda_parse_path(parameter_path, self.snudda_data)
        elif self.neuron_path:
            self.parameter_path = snudda_parse_path(os.path.join(self.neuron_path, "parameters.json"), self.snudda_data)
        else:
            self.parameter_path = None

        if meta_path:
            self.meta_path = snudda_parse_path(meta_path, self.snudda_data)
        elif self.neuron_path:
            self.meta_path = snudda_parse_path(os.path.join(self.neuron_path, "meta.json"), self.snudda_data)
        else:
            self.meta_path = None

        if modulation_path:
            self.modulation_path = snudda_parse_path(modulation_path, self.snudda_data)
        elif self.neuron_path:
            self.modulation_path = snudda_parse_path(os.path.join(self.neuron_path, "modulation.json"), self.snudda_data)

            if not os.path.exists(self.modulation_path):
                self.modulation_path = None
        else:
            self.modulation_path = None

        if isinstance(reaction_diffusion_path, dict):
            self.reaction_diffusion_path = reaction_diffusion_path

        elif reaction_diffusion_path:
            self.reaction_diffusion_path = snudda_parse_path(reaction_diffusion_path, self.snudda_data)

        elif self.neuron_path:
            self.reaction_diffusion_path = snudda_parse_path(os.path.join(self.neuron_path, "reaction_diffusion.json"),
                                                             self.snudda_data)

            if not os.path.exists(self.reaction_diffusion_path):
                self.reaction_diffusion_path = None
        else:
            self.reaction_diffusion_path = None

        self.neuron_name = neuron_name
        self.parameter_info = None
        self.meta_info = None
        self.modulation_info = None
        self.virtual_neuron = virtual_neuron
        self.load_morphology = load_morphology

        self.morphology_cache = dict()
        self.morphology_lookup = dict()

        self.load_info()

    def load_info(self):
        """ Reads information about the neuron prototype. """

        self.morphology_cache = dict()
        self.morphology_lookup = dict()

        par_path = self.parameter_path


        if self.meta_path and os.path.exists(self.meta_path):
            try:
                with open(self.meta_path, "r") as fm:
                    self.meta_info = json.load(fm)
            except Exception as e:
                print(f"!! Attempting to load {self.meta_path}")
                import traceback
                print(traceback.format_exc())
                raise e

        if par_path is None or not os.path.exists(par_path):
            if self.verbose:
                print(f"Missing parameters.json : {par_path}")

            self.parameter_info = None
            return

        with open(par_path, "r") as f:
            self.parameter_info = json.load(f)

        if self.meta_info and isinstance(self.parameter_info, dict):
            # Check that all parameter_keys are in meta, otherwise remove them
            for par_key in list(self.parameter_info.keys()):
                if par_key not in self.meta_info:
                    del self.parameter_info[par_key]
                    if self.verbose:
                        print(f"Parameter key {par_key} missing in {self.meta_path}")

        # We now expect a dictionary of parameter sets. If it is a list, we convert it to a dictionary
        if type(self.parameter_info) == list:
            self.parameter_info = {"default": self.parameter_info}

        mod_path = self.modulation_path

        if mod_path is not None and os.path.exists(mod_path):
            with open(mod_path, "r") as f:
                self.modulation_info = json.load(f)
        else:
            self.modulation_info = None

    def get_num_morphologies(self, parameter_id=None, parameter_key=None):

        if parameter_key is None:
            parameter_key = self.get_parameter_key(parameter_id=parameter_id)
        elif parameter_id is not None:
            assert parameter_key == self.get_parameter_key(parameter_id=parameter_id), \
                f"Internal error, parameter_id {parameter_id} does not match parameter_key {parameter_key} specified."

        if self.meta_info and parameter_key:
            morph_key_list = self.meta_info[parameter_key].keys()
            return len(morph_key_list)
        else:
            return 1

    def get_parameter_key(self, parameter_id):

        if parameter_id is None:
            return None

        # assert parameter_id is not None, (f"get_parameter_key: parameter_id is None, for {self.neuron_name}"
        #                                   f"\nNeuron path: {self.neuron_path}")

        if self.parameter_info:
            par_key_list = sorted(list(self.parameter_info.keys()))
            par_key = par_key_list[parameter_id % len(par_key_list)]
        elif self.meta_info:
            if self.verbose:
                print(f"Missing parameter file for {self.parameter_path}, using meta.json for parameter key list.")
            par_key_list = sorted(list(self.meta_info.keys()))
            par_key = par_key_list[parameter_id % len(par_key_list)]
        else:
            par_key = None

        return par_key

    def get_morph_key(self,  morphology_id, parameter_id=None, parameter_key=None):

        if parameter_id is None:
            par_key = parameter_key
        else:
            par_key = self.get_parameter_key(parameter_id=parameter_id)
            assert parameter_key is None or par_key == parameter_key, \
                (f"parameter_id = {parameter_id} gives parameter_key {par_key}, "
                 f"different from parameter_key {parameter_key} provided")

        if self.meta_info:
            assert par_key in self.meta_info, f"Parameter key {par_key} missing in {self.meta_path}"
            morph_key_list = sorted(list(self.meta_info[par_key].keys()))
            assert len(morph_key_list) > 0, f"No morphologies available for parameter key {par_key} in {self.meta_path}"
            morph_key = morph_key_list[morphology_id % len(morph_key_list)]
        else:
            morph_key = None

        return morph_key

    def get_modulation_key(self, modulation_id,
                           parameter_id=None, morphology_id=None,
                           parameter_key=None, morphology_key=None):

        if modulation_id is None:
            return None

        if self.modulation_info:
            if type(self.modulation_info) == list:
                # Old format without keys
                return None

            if parameter_key is None:
                parameter_key = self.get_parameter_key(parameter_id)

            if morphology_key is None:
                morphology_key = self.get_morph_key(parameter_id=None, parameter_key=parameter_key,
                                                    morphology_id=morphology_id)

            modulation_key_list = self.meta_info[parameter_key][morphology_key]["neuromodulation"]

            if len(modulation_key_list) == 0:
                print("Modulation key list empty")
                print(f"Path: {self.neuron_path}, param: {parameter_key}, morph: {morphology_key}")
                import pdb
                pdb.set_trace()

            # modulation_key_list = list(self.modulation_info.keys())
            modulation_key = modulation_key_list[modulation_id % len(modulation_key_list)]
        else:
            modulation_key = None

        return modulation_key

    def get_input_parameters(self, parameter_id=None, morphology_id=None, parameter_key=None, morphology_key=None):

        if parameter_key is not None and morphology_key is not None:
            if self.meta_info and "input" in self.meta_info[parameter_key][morphology_key]:
                input_info = self.meta_info[parameter_key][morphology_key]["input"]
            else:
                input_info = dict()

        else:
            # Fallback if the user gave ID instead of Keys (will remove this at some point)
            par_key = self.get_parameter_key(parameter_id=parameter_id)
            morph_key = self.get_morph_key(parameter_id=parameter_id, morphology_id=morphology_id)

            if self.meta_info and "input" in self.meta_info[par_key][morph_key]:
                input_info = self.meta_info[par_key][morph_key]["input"]
            else:
                input_info = dict()

        return input_info

    def get_morphology(self, parameter_key=None, morphology_key=None, parameter_id=None, morphology_id=None):

        """
        Returns morphology for a given parameter_id, morphology_id combination.
        (Each parameter set has a set of morphologies that it is valid for)
        """

        # TODO: In the future remove parameter_id and morphology_id and only use keys
        if parameter_id is not None:
            par_key = self.get_parameter_key(parameter_id=parameter_id)
            if parameter_key:
                assert par_key == parameter_key, \
                    f"Mismatch. Parameter ID {parameter_id} has parameter_key {par_key}, " \
                    f"not {parameter_key} (Is snudda_data set correctly?)"
        elif parameter_key:
            par_key = parameter_key
        else:
            par_key = None

        if self.meta_info and par_key is not None:

            if morphology_id is not None:
                morph_key = self.get_morph_key(parameter_id=parameter_id, morphology_id=morphology_id,
                                               parameter_key=par_key)
                if morphology_key:
                    if morph_key != morphology_key:
                        raise KeyError(f"Mismatch: Expected morphology_key {morph_key}, got {morphology_key}")

            elif morphology_key is not None:
                morph_key = morphology_key
                if morphology_key not in self.meta_info[par_key]:
                    raise KeyError(f"Missing morphology_key {morphology_key}")
            else:
                raise KeyError(f"morphology_id or morphology_key must be specified if meta_info is used")

            morph_path = os.path.join(self.morphology_path, self.meta_info[par_key][morph_key]["morphology"])
            assert os.path.isfile(morph_path), f"Morphology file {morph_path} is missing (listed in {self.meta_path})"

        elif self.morphology_path and os.path.isfile(self.morphology_path):
            # Fallback if morphology file is specified in the path
            morph_path = self.morphology_path
            morph_key = morphology_key
        else:
            # No morphology
            file_list = glob.glob(os.path.join(self.neuron_path, "*swc"))
            morph_key = morphology_key

            if len(file_list) > 0:
                morph_path = file_list[0]

                assert len(file_list) == 1, f"More than one morphology available in {self.neuron_path}"

            else:
                print(f"No morphology in {self.neuron_path}")
                morph_path = None

        if morph_path is None:
            print(f"morph_path is None for {self.neuron_name} path: {self.neuron_path}. "
                  f"Is SNUDDA_DATA set correctly? ({os.environ['SNUDDA_DATA'] if 'SNUDDA_DATA' in os.environ else None})")
            import pdb
            pdb.set_trace()

        assert morph_path is not None, \
            (f"morph_path is None for {self.neuron_name} path: {self.neuron_path} (neuron_path should be a directory, not a SWC file). "
             f"Is SNUDDA_DATA set correctly? ({os.environ['SNUDDA_DATA']})")

        return morph_path, morph_key

    def get_parameters(self, parameter_key=None, parameter_id=None):
        """
        Returns neuron parameter set

        Args:
            parameter_key (str) : parameter key for the parameter set
            parameter_id (int) : parameter ID, which of the parameter sets to use

        Returns:
            One parameter set as a list of dictionaries
        """

        if self.parameter_info:
            if parameter_key is not None:
                assert parameter_key in self.parameter_info, \
                    f"Parameter key {parameter_key} not in parameter_info {self.parameter_info}. ERROR, no key: {parameter_key}"
                par_set = self.parameter_info[parameter_key]
            else:
                assert parameter_id is not None, "If parameter_key is not set, then parameter_id must be set"
                par_key_list = list(self.parameter_info.keys())
                par_key = par_key_list[parameter_id % len(par_key_list)]
                par_set = self.parameter_info[par_key]
        else:
            par_set = []

        return par_set

    def get_modulation_parameters(self, modulation_key=None, modulation_id=None):

        """
        Returns modulation parameter set, give either ID or key

        Args:
            modulation_key (str) : modulation key
            modulation_id (int) : modulation ID, which of the modulation parameter sets to use

        Returns:
            One parameter set as a list of dictionaries
        """

        if self.modulation_info:
            if modulation_key:
                modulation_set = self.modulation_info[modulation_key]
            elif modulation_id is not None:
                possible_keys = [x for x in self.modulation_info.keys()]
                modulation_set = self.modulation_info[possible_keys[modulation_id % len(self.modulation_info)]]
            else:
                assert False, f"You need to specify modulation_key or modulation_id to look up a modulation set"
        else:
            modulation_set = []

        return modulation_set

    def instantiate(self):
        """
        Instantiates all morphologies at once, instead of on demand.
        """

        n_par = len(self.parameter_info) if self.parameter_info is not None else 1

        for par_id in range(0, n_par):
            if self.verbose:
                print(f"Instantiates par_id = {par_id}")

            par_data = self.get_parameters(parameter_id=par_id)
            n_morph = self.get_num_morphologies(parameter_id=par_id)

            for morph_id in range(0, n_morph):
                morph_path, morph_key = self.get_morphology(parameter_id=par_id, morphology_id=morph_id)
                morph_tag = os.path.basename(morph_path)

                if morph_tag not in self.morphology_cache:
                    if self.verbose:
                        print(f"morph_tag = {morph_tag}")

                    self.morphology_cache[morph_tag] = NeuronMorphologyExtended(swc_filename=morph_path,
                                                                                param_data=self.parameter_path,
                                                                                mech_filename=self.mechanism_path,
                                                                                reaction_diffusion=self.reaction_diffusion_path,
                                                                                neuron_path=self.neuron_path,
                                                                                snudda_data=self.snudda_data,
                                                                                name=self.neuron_name,
                                                                                morphology_key=morph_key,
                                                                                load_morphology=self.load_morphology,
                                                                                virtual_neuron=self.virtual_neuron)

    def apply(self, function_name, arguments):
        """
        Applies function to all morphology prototypes
        """

        res = []

        for m in self.morphology_cache.values():
            function = getattr(m, function_name)
            res.append(function(*arguments))

        return res

    def all_have_axon(self):

        all_axon = True

        for m in self.morphology_cache.values():
            if 2 not in m.morphology_data["neuron"].sections \
              or len(m.morphology_data["neuron"].sections[2]) == 0:
                all_axon = False

        return all_axon

    def all_have_dend(self):

        all_dend = True

        for m in self.morphology_cache.values():
            if 3 not in m.morphology_data["neuron"].sections \
              or len(m.morphology_data["neuron"].sections[3]) == 0:
                all_dend = False

        return all_dend

    def clone(self, parameter_id=None, morphology_id=None, modulation_id=None,
              parameter_key=None, morphology_key=None, modulation_key=None,
              position=None, rotation=None, get_cache_original=False,
              morphology_path=None):
        """
        Creates a clone of the neuron prototype, with given position and rotation.

        Use either keys or id (keys have precedence if both given)

        Args:
            parameter_id (int) : parameter set to use
            morphology_id (int) : morphology set to use, a parameter set can have multiple morphology variations
            modulation_id (int) : neuro-modulation parameter set to use

            parameter_key (str) : parameter key for parameter set in parameters.json
            morphology_key (str) : morphology key for morphology in meta.json
            modulation_key (str) : modulation key for neuromodulation in modulation.json

            position (float,float,float) : position of neuron clone
            rotation : rotation (3x3 numpy matrix)
            get_cache_original (bool) : return the original rather than a clone

            morphology_path : If morphology path is given it can override morphology key
        """

        if morphology_path is not None:
            morph_path = morphology_path
        else:
            morph_path, morph_key = self.get_morphology(parameter_key=parameter_key, morphology_key=morphology_key,
                                                        parameter_id=parameter_id, morphology_id=morphology_id)

            if morphology_key is None:
                morphology_key = morph_key
            else:
                assert morphology_key == morph_key, \
                    f"Internal mismatch requested morphology_key {morphology_key}, got {morph_key}"

        morph_tag = os.path.basename(morph_path)

        if morph_tag not in self.morphology_cache:

            # TODO: hoc file will depend on both morphology_id and parameter_id, we ignore it for now
            self.morphology_cache[morph_tag] = NeuronMorphologyExtended(name=self.neuron_name,
                                                                        position=None,  # This is set further down when using clone
                                                                        rotation=None,
                                                                        swc_filename=morph_path,
                                                                        snudda_data=self.snudda_data,
                                                                        param_data=self.parameter_path,
                                                                        mech_filename=self.mechanism_path,
                                                                        reaction_diffusion=self.reaction_diffusion_path,
                                                                        neuron_path=self.neuron_path,
                                                                        parameter_key=parameter_key,
                                                                        morphology_key=morphology_key,
                                                                        modulation_key=modulation_key,
                                                                        load_morphology=self.load_morphology,
                                                                        virtual_neuron=self.virtual_neuron,
                                                                        verbose=self.verbose)


        if get_cache_original:
            assert position is None and rotation is None and modulation_id is None, \
                "If get_cache_original is passed position, rotation and modulation_id must be None"
            morph = self.morphology_cache[morph_tag]
        else:
            # Add the keys also for parameter, morphology and modulation
            if parameter_key is None:
                parameter_key = self.get_parameter_key(parameter_id=parameter_id)

            if morphology_key is None:
                morphology_key = self.get_morph_key(parameter_key=parameter_key, morphology_id=morphology_id)

            if modulation_key is None:
                modulation_key = self.get_modulation_key(parameter_key=parameter_key,
                                                         morphology_key=morphology_key,
                                                         modulation_id=modulation_id)

            morph = self.morphology_cache[morph_tag].clone(position=position,
                                                           rotation=rotation,
                                                           parameter_key=parameter_key,
                                                           morphology_key=morphology_key,
                                                           modulation_key=modulation_key)

            if self.meta_info is not None and "axon_density" in self.meta_info.get(parameter_key, {}).get(morphology_key, {}):
                axon_density_type, axon_density, axon_density_bounds = self.meta_info[parameter_key][morphology_key]["axon_density"]

                if axon_density_type == "r":
                    morph.set_axon_voxel_radial_density(density=axon_density,
                                                        max_axon_radius=axon_density_bounds)
                elif axon_density_type == "xyz":
                    morph.set_axon_voxel_xyz_density(density=axon_density,
                                                     axon_density_bounds_xyz=axon_density_bounds)
                else:
                    raise ValueError(f"Unknown axon_density type {axon_density_type}, {self.neuron_path =}")

        return morph



import unittest
import os
import numpy as np

from snudda.utils.ablate_network import SnuddaAblateNetwork
from snudda.utils.load import SnuddaLoad
from snudda.analyse import SnuddaAnalyse

#
# This unit test uses a pre-generated network, and checks that neurons and connections in it are correctly removed
#
# To regenerate the network, stand in the Snudda/tests directory and do:
#
#    snudda init networks/modify_network --size 200
#    snudda place networks/modify_network
#    snudda detect networks/modify_network
#    snudda prune networks/modify_network


class SnuddaModifyNetworkTestCase(unittest.TestCase):

    def setUp(self) -> None:

        if os.path.dirname(__file__):
            os.chdir(os.path.dirname(__file__))

        self.network_path = os.path.join("networks", "modify_network")
        self.original_file = os.path.join(self.network_path, "network-synapses.hdf5")

        if not os.path.exists(self.original_file):
            print("Missing test network, generating it first.")
            from snudda.init import SnuddaInit
            SnuddaInit(network_path=self.network_path, struct_def={"Striatum": 200}, random_seed=999)

            from snudda.place import SnuddaPlace
            sp = SnuddaPlace(network_path=self.network_path)
            sp.place()

            from snudda.detect import SnuddaDetect, SnuddaPrune
            sd = SnuddaDetect(network_path=self.network_path)
            sd.detect()

            spp = SnuddaPrune(network_path=self.network_path)
            spp.prune()

        self.snudda_load_original = SnuddaLoad(network_file=self.original_file, load_synapses=True)
        self.original_type_count = self.get_number_of_type(self.snudda_load_original.data)

    def test_remove_neurons(self):

        """ Verify that neuron type is removed correctly """

        profile_run = False

        new_file_dspn = os.path.join(self.network_path, "dspn-neurons-removed.hdf5")
        new_file_ispn = os.path.join(self.network_path, "ispn-neurons-removed.hdf5")
        new_file_ispn_dspn = os.path.join(self.network_path, "ispn-dspn-connections-removed.hdf5")

        mod_network = SnuddaAblateNetwork(network_file=self.original_file)
        sa_orig = SnuddaAnalyse(hdf5_file=self.original_file)
        dspn_id_orig = self.snudda_load_original.get_neuron_id_of_type(neuron_type="dSPN")
        ispn_id_orig = self.snudda_load_original.get_neuron_id_of_type(neuron_type="iSPN")

        with self.subTest(msg="Testing removal of iSPN to dSPN synapses"):
            mod_network.reset_network()
            mod_network.remove_connection(pre_neuron_type="iSPN", post_neuron_type="dSPN")
            mod_network.write_network(out_file_name=new_file_ispn_dspn)

            sa_new = SnuddaAnalyse(hdf5_file=new_file_ispn_dspn)

            print("Comparing connection matrices")

            ref_mat = sa_orig.connection_matrix.copy()
            for i_id in ispn_id_orig:
                ref_mat[i_id, dspn_id_orig] = 0

            # When all ispn to dspn connections are removed in ref matrix, it should match the new connection matrix
            print("Checking that new connection matrix match reference")
            self.assertTrue((sa_new.connection_matrix != ref_mat).nnz == 0)

        with self.subTest(msg="Testing removal of neuron type dSPN"):
            mod_network.reset_network()
            mod_network.remove_neuron_type(neuron_type="dSPN")

            if profile_run:
                import cProfile
                prof_file = "profiling-info.prof"
                cProfile.runctx("mod_network.write_network(out_file_name=new_file_dspn)",
                                None, locals(), filename=prof_file)

                # To analyse profile data:
                import pstats
                from pstats import SortKey
                p = pstats.Stats(prof_file)
                p.strip_dirs().sort_stats(SortKey.CUMULATIVE).print_stats(100)

            else:
                mod_network.write_network(out_file_name=new_file_dspn)

            snudda_load_new = SnuddaLoad(network_file=new_file_dspn, load_synapses=True)
            new_type_count = self.get_number_of_type(snudda_load_new.data)

            for neuron_type in self.original_type_count:
                if neuron_type == "dSPN":
                    self.assertTrue(neuron_type not in new_type_count)
                else:
                    self.assertEqual(new_type_count[neuron_type], self.original_type_count[neuron_type])

            sa_new = SnuddaAnalyse(hdf5_file=new_file_dspn)
            # Remove dSPN from original connection matrix and check it matches with new connection matrix
            keep_idx = list(set(np.arange(0, len(self.snudda_load_original.data["neurons"]))) - set(dspn_id_orig))
            ref_mat = sa_orig.connection_matrix.copy()[keep_idx, :][:, keep_idx]

            self.assertTrue((sa_new.connection_matrix != ref_mat).nnz == 0)

        with self.subTest(msg="Testing removal of neuron type iSPN, p=0.3"):
            mod_network.reset_network()
            mod_network.remove_neuron_type(neuron_type="iSPN", p_remove=0.3)
            mod_network.write_network(out_file_name=new_file_ispn)

            snudda_load_new = SnuddaLoad(network_file=new_file_ispn, load_synapses=True)
            new_type_count = self.get_number_of_type(snudda_load_new.data)

            for neuron_type in self.original_type_count:
                if neuron_type == "iSPN":
                    # Check that we are within a range of 30% removed
                    self.assertTrue(0.5 * self.original_type_count[neuron_type]
                                    < new_type_count[neuron_type]
                                    < 0.9 * self.original_type_count[neuron_type])
                else:
                    self.assertEqual(new_type_count[neuron_type], self.original_type_count[neuron_type])

    def get_number_of_type(self, snudda_load_data):

        type_counting = dict()

        neuron_types = [n["type"] for n in snudda_load_data["neurons"]]
        unique_neuron_types = set(neuron_types)

        for unt in unique_neuron_types:
            type_counting[unt] = np.sum([nt == unt for nt in neuron_types])

        return type_counting

        self.assertEqual(True, False)  # add assertion here


if __name__ == '__main__':
    unittest.main()
